unit LDXGraphics;

interface

uses
	Windows, Messages, SysUtils, Math, SysConst,
	Direct3D8, D3DX8,
	SysHacks,
	LObjects, LWindows;

{ Alias }

type
	ITexture = IDirect3DTexture8;
	TTextureLock = TD3DLockedRect;
	TMaterial = TD3DMaterial8;
	TMatrix = TD3DMatrix;
	TVector = TD3DVector;
	IVertexBuffer = IDirect3DVertexBuffer8;

	TD3DColor = Direct3D8.TD3DColor;

const
	IdentityMatrix : TD3DMatrix = (
		_11: 1; _12: 0; _13: 0; _14: 0;
		_21: 0; _22: 1; _23: 0; _24: 0;
		_31: 0; _32: 0; _33: 1; _34: 0;
		_41: 0; _42: 0; _43: 0; _44: 1 );
	ZeroMatrix : TD3DMatrix = (
		_11: 0; _12: 0; _13: 0; _14: 0;
		_21: 0; _22: 0; _23: 0; _24: 0;
		_31: 0; _32: 0; _33: 0; _34: 0;
		_41: 0; _42: 0; _43: 0; _44: 0 );

type
	TTextureFormat = D3DFORMAT;

const
	tfA1R5G5B5 = D3DFMT_A1R5G5B5;
	tfA8R8G8B8 = D3DFMT_A8R8G8B8;

type
	TPrimitiveType = D3DPRIMITIVETYPE;

const
	ptTriangleStrip = D3DPT_TRIANGLESTRIP;

type
	TRenderState = D3DRENDERSTATETYPE;

const
	rsFogEnable = D3DRS_FOGENABLE;
	rsSpecularEnable = D3DRS_SPECULARENABLE;

type
	TLightType = D3DLIGHTTYPE;

const
	ltDirectional = D3DLIGHT_DIRECTIONAL;

type
	TTextureLockMode = (
		tlNoSysLock = D3DLOCK_NOSYSLOCK,
		_tlHigh = High(Integer)
	);

{ Vertex }

const
	D3DFVF_VERTEX = D3DFVF_XYZ or D3DFVF_NORMAL or D3DFVF_TEX1;
	D3DFVF_VERTEX_2D = D3DFVF_XYZRHW or D3DFVF_DIFFUSE or D3DFVF_SPECULAR or D3DFVF_TEX1;

type
	TVertex = record
		x, y, z: Single;
	case Boolean of
	False: (
		nx, ny, nz: Single;
		tu, tv: Single;
	);
	True: (
		rhw: Single;
		diffuse, specular: DWORD;
	);
	end;

{ Screen }

type
	T3DScreen = class;

	IScreenCreator = interface(IInterface)
		procedure InitializeScreen;
		procedure FinalizeScreen;
	end;

	PRendering = ^TRendering;
	TRendering = record
		Render: TMethod;
		World: TD3DMatrix;
		Screen: T3DScreen;
		z, absx: Single;
		Blending: Boolean;
		Data: Pointer;
		Before, After: Integer; {binary tree}
	end;

	TRender = procedure(const Rendering: TRendering) of object;

	TLight = class(T0CollectionItem)
	private
		FData: TD3DLight8;
		FID: Cardinal;
		procedure SetEnabled(const Value: Boolean);
		procedure SetAmbient(const Value: TD3DColorValue);
		procedure SetDiffuse(const Value: TD3DColorValue);
		procedure SetSpecular(const Value: TD3DColorValue);
		procedure SetDirection(const Value: TD3DVector);
		function GetEnabled: Boolean;
	protected
		procedure Modify;
	public
		constructor Create(Screen: T3DScreen; AType: TLightType);
		procedure BeforeDestruction; override;
		property Enabled: Boolean read GetEnabled write SetEnabled;
		property ID: Cardinal read FID;
		property Specular: TD3DColorValue read FData.Specular write SetSpecular;
		property Ambient: TD3DColorValue read FData.Ambient write SetAmbient;
		property Diffuse: TD3DColorValue read FData.Diffuse write SetDiffuse;
		property Direction: TD3DVector read FData.Direction write SetDirection;
	end;

	TLights = class(T0Collection)
	private
		FUsedID: TIntegerSet;
		FModified: TIntegerSet;
		FEnabled: TIntegerSet;
		procedure SetAmbient(const Value: TD3DCOLOR);
	public
		property Ambient: TD3DCOLOR write SetAmbient;
	end;

	TVertex4Array = array[0..3] of TVertex;

	T3DScreen = class(T0Window)
	private
		FFullScreenSize: TSize;
		FDirect3D: IDirect3D8;
		FDirect3DDevice: IDirect3DDevice8;
		FCreator: IScreenCreator;
		FScreenWindow: HWND;
		FLights: TLights;
		FRenderingObjects: array of TRendering;
		FFullScreen: Boolean;
		FStandby: Boolean;
		FCursorVisible: Boolean;
		FZBuffer: Boolean;
		procedure WMSetFocus(var Message); message WM_SETFOCUS;
		procedure WMLightWindows(var Message: TMessage); message WM_LIGHTWINDOWS;
		procedure SetFullScreen(const Value: Boolean);
		function GetProjectionMatrix: TD3DMatrix;
		procedure SetProjectionMatrix(const Value: TD3DMatrix);
		function GetViewMatrix: TD3DMatrix;
		procedure SetViewMatrix(const Value: TD3DMatrix);
		function GetMaterial: TD3DMaterial8;
		procedure SetMaterial(const Value: TD3DMaterial8);
		procedure SetTextures(Index: Integer; const Value: IDirect3DTexture8);
		procedure SetTexture(const Value: IDirect3DTexture8);
		procedure SetFog(const Value: Boolean);
		function GetRenderState(Index: TRenderState): Cardinal;
		procedure SetRenderState(Index: TRenderState; const Value: Cardinal);
	protected
		procedure Initialize; override;
		procedure Finalize; override;
		procedure InitializeDirectXGraphics;
		procedure FinalizeDirectXGraphics;
		property ScreenWindow: HWND read FScreenWindow;
	public
		constructor Create(AParent: TAnyWindow; const APosition: TPoint; const ASize: TSize;
						const ACreator: IScreenCreator; AFullScreen: Boolean;
						const AFullScreenSize: TSize;
						AZBuffer: Boolean = True);
		procedure Realign; override;
		procedure Clear(Color: TD3DColor);
		procedure ClearZBuffer;
		procedure BeginScene;
		procedure EndScene;
		procedure Flush;
		procedure Present;
		procedure ReadyLights;
		procedure ReadyStatesForNormal;
		procedure ReadyStatesForAlpha;
		procedure ReadyStatesForAdd;
		procedure ReadyStatesForMultiply;
		procedure ReadyStatesForSubtract;
		procedure ReadyFog(ANear, AFar: Single; Color: TD3DColor);
		procedure Render(const ARender: TRender; const AWorld: TD3DMatrix; ABlending: Boolean = False; AData: Pointer = nil);
		procedure ClearTexture;
		procedure Draw(const AVerteces: IDirect3DVertexBuffer8; AType: TPrimitiveType; Start, Count: Cardinal); overload;
		procedure Draw(const AVerteces: TVertex4Array); overload;
		function ScreenSize: TSize;
		function ScreenCursorPos: TPoint;
		function OptimalProjectionMatrix: TD3DMatrix;
		function GetBackBuffer: IDirect3DSurface8;
		function CreateVertexBuffer(const V: array of TVertex): IDirect3DVertexBuffer8;
		function CreateTexture(W, H: Integer): IDirect3DTexture8;
		property FullScreen: Boolean read FFullScreen write SetFullScreen;
		property Standby: Boolean read FStandby;
		property Direct3D: IDirect3D8 read FDirect3D;
		property Direct3DDevice: IDirect3DDevice8 read FDirect3DDevice;
		property ProjectionMatrix: TD3DMatrix read GetProjectionMatrix write SetProjectionMatrix;
		property ViewMatrix: TD3DMatrix read GetViewMatrix write SetViewMatrix;
		property Lights: TLights read FLights;
		property CursorVisible: Boolean read FCursorVisible write FCursorVisible;
		property Material: TD3DMaterial8 read GetMaterial write SetMaterial;
		property Texture: IDirect3DTexture8 write SetTexture;
		property Textures[Index: Integer]: IDirect3DTexture8 write SetTextures;
		property Fog: Boolean write SetFog;
		property FullScreenSize: TSize read FFullScreenSize;
		property RenderState[Index: TRenderState]: Cardinal read GetRenderState write SetRenderState;
	end;

function D3DColorValue(const R, G, B: Single; const A: Single = 1): TD3DColorValue;
function D3DColor(const R, G, B: Single; const A: Single = 1): TD3DColor;

function ProjectionMatrix(const FOV, Aspect, NearPlane, FarPlane: Single): TD3DMatrix;
function ViewMatrix(const From, At, WorldUp: TD3DVector): TD3DMatrix;
function Billboard(const World, View: TD3DMatrix): TD3DMatrix;
function TranslateMatrix(X, Y, Z: Single): TD3DMATRIX;
function RotateXMatrix(X: Single): TD3DMATRIX;
function RotateYMatrix(Y: Single): TD3DMATRIX;
function RotateZMatrix(Z: Single): TD3DMATRIX;
function ScaleMatrix(X, Y, Z: Single): TMatrix;
function VectorMulMatrix(const V: TD3DVector; const M: TD3DMatrix): TD3DVector;
function MatrixMulMatrix(const A, B: TD3DMatrix): TD3DMatrix;
function Vector(x, y, z: Single): TVector;
function VectorAdd(const v1, v2: TD3DVector) : TD3DVector;

function XYRectangle(const X1, Y1, X2, Y2: Single): TVertex4Array;
function YZRectangle(const Y1, Z1, Y2, Z2: Single; const X: Single = 0): TVertex4Array;
function ZXRectangle(const Z1, X1, Z2, X2: Single; const Y: Single = 0): TVertex4Array;
function Rectangle(const X1, Y1, X2, Y2: Integer; const U1, V1, U2, V2: Single): TVertex4Array;

function ColorMaterial(const Ambient: TD3DColorValue): TD3DMaterial8;
procedure SetDeviceCursor(const Device: IDirect3DDevice8; Cursor: HCURSOR);

{ D3DX }

function LoadTexture(
	const Device: IDirect3DDevice8;
	const FileName: WideString;
	Format: TTextureFormat;
	TransparentColor: TD3DColor): ITexture; overload;
function LoadTexture(
	const Device: IDirect3DDevice8;
	P: Pointer; Size: Cardinal;
	Format: TTextureFormat;
	TransparentColor: TD3DColor): ITexture; overload;

const
	Size640x480: TSize = (cx: 640; cy: 480);
	Size800x600: TSize = (cx: 800; cy: 600);

implementation

function SetProjectionMatrix(var mat: TD3DMatrix; fFOV, fAspect, fNearPlane, fFarPlane: Single): HResult;
var
  w, h, q: Single;
begin
  if (abs(fFarPlane-fNearPlane) < 0.01) or (abs(sin(fFOV/2)) < 0.01) then
  begin
    Result := E_INVALIDARG;
    Exit;
  end;
  h := (cos(fFOV/2)/sin(fFOV/2));
  w := fAspect * h;
  Q := fFarPlane / (fFarPlane - fNearPlane);
  ZeroMemory(@mat, SizeOf(mat));
  mat._11 := w;
  mat._22 := h;
  mat._33 := Q;
  mat._34 := 1.0;
  mat._43 := -Q*fNearPlane;
  Result := S_OK;
end;

function D3DColorValue(const R, G, B, A: Single): TD3DColorValue;
begin
	Result.r := R;
	Result.g := G;
	Result.b := B;
	Result.a := A;
end;

function D3DColor(const R, G, B, A: Single): TD3DColor;
begin
	Result := Round(A * 255) shl 24
		or Round(R * 255) shl 16
		or Round(G * 255) shl 8
		or Round(B * 255)
end;

function ProjectionMatrix(const FOV, Aspect, NearPlane, FarPlane: Single): TD3DMatrix;
begin
	//D3DXMatrixPerspectiveFovLH(Result, FOV, Aspect, NearPlane, FarPlane)
	SetProjectionMatrix(Result, FOV, Aspect, NearPlane, FarPlane)
end;

function ViewMatrix(const From, At, WorldUp: TD3DVector): TD3DMatrix;
begin
	D3DXMatrixLookAtLH(Result, From, At, WorldUp)
	//SetViewMatrix(Result, From, At, WorldUp)
end;

function Billboard(const World, View: TD3DMatrix): TD3DMatrix;
begin
	Result := World;
	Result._11 := View._11; Result._12 := View._21; Result._13 := View._31;
	Result._21 := View._12; Result._22 := View._22; Result._23 := View._32;
	Result._31 := View._13; Result._32 := View._23; Result._33 := View._33;
end;

function TranslateMatrix(X, Y, Z: Single): TD3DMATRIX;
begin
	Result := IdentityMatrix;
	Result._41 := X; Result._42 := Y; Result._43 := Z;
end;

function RotateXMatrix(X: Single): TD3DMATRIX;
begin
	D3DXMatrixRotationX(Result, X);
end;

function RotateYMatrix(Y: Single): TD3DMATRIX;
begin
	D3DXMatrixRotationY(Result, Y);
end;

function RotateZMatrix(Z: Single): TD3DMATRIX;
begin
	D3DXMatrixRotationZ(Result, Z);
end;

function ScaleMatrix(X, Y, Z: Single): TMatrix;
begin
	D3DXMatrixScaling(Result, X, Y, Z);
end;

function VectorMulMatrix(const V: TD3DVector; const M: TD3DMatrix): TD3DVector;
var
	x, y, z, w: Single;
begin
	x := v.x*m._11 + v.y*m._21 + v.z* m._31 + m._41;
	y := v.x*m._12 + v.y*m._22 + v.z* m._32 + m._42;
	z := v.x*m._13 + v.y*m._23 + v.z* m._33 + m._43;
	w := v.x*m._14 + v.y*m._24 + v.z* m._34 + m._44;

	if abs(w) < 1.0e-5 then
		raise EUnderflow.Create(SUnderflow);

	Result.x := x/w;
	Result.y := y/w;
	Result.z := z/w;
end;

function MatrixMulMatrix(const A, B: TD3DMatrix): TD3DMatrix;
begin
	D3DXMatrixMultiply(Result, B, A)
end;

function Vector(x, y, z: Single): TVector;
begin
	Result.x := x;
	Result.y := y;
	Result.z := z;
end;

function VectorAdd(const v1, v2: TD3DVector) : TD3DVector;
begin
	result.x := v1.x+v2.x;
	result.y := v1.y+v2.y;
	result.z := v1.z+v2.z;
end;

procedure SetDeviceCursor(const Device: IDirect3DDevice8; Cursor: HCURSOR);
type
	PCOLORREFArray = ^TCOLORREFArray;
	TCOLORREFArray = array[0..32767] of COLORREF;
var
	CursorInfo: ICONINFO;
	ColorDC, MaskDC, ScreenDC: HDC;
	CursorBitmap: BITMAP;
	CursorBitmapInfo: BITMAPINFO;
	IsMonochrome: Boolean;
	Width, SourceHeight, Height: Integer;
	Color, Mask: COLORREF;
	X, Y: Integer;
	HR: HRESULT;
	p, MaskBuffer, ColorBuffer: PCOLORREFArray;
	Old: HGDIOBJ;
	Locked: TD3DLOCKED_RECT;
	Result: IDirect3DSurface8;
begin
	MaskBuffer := nil;
	ColorBuffer := nil;

	FillChar(CursorInfo, SizeOf(CursorInfo), 0);

	if not GetIconInfo(Cursor, CursorInfo) then
		RaiseLastOSError;

	try

		if GetObjectW(CursorInfo.hbmMask, SizeOf(CursorBitmap), @CursorBitmap) = 0 then
			RaiseLastOSError;

		Width := CursorBitmap.bmWidth;
		SourceHeight := CursorBitmap.bmHeight;
		Height := SourceHeight;
		IsMonochrome := CursorInfo.hbmColor = 0;
		if IsMonochrome then
			Height := Height div 2;

		HR := Device.CreateImageSurface(Width, Height, D3DFMT_A8R8G8B8, Result);
		if FAILED(HR) then
			raise EOSError.Create(D3DXGetErrorString(HR));

		FillChar(CursorBitmapInfo, sizeof(CursorBitmapInfo), 0);
		CursorBitmapInfo.bmiHeader.biSize := Sizeof(CursorBitmapInfo.bmiHeader);
		CursorBitmapInfo.bmiHeader.biWidth := Width;
		CursorBitmapInfo.bmiHeader.biHeight := SourceHeight;
		CursorBitmapInfo.bmiHeader.biPlanes := 1;
		CursorBitmapInfo.bmiHeader.biBitCount := 32;
		CursorBitmapInfo.bmiHeader.biCompression := BI_RGB;

		ScreenDC := GetDC(0);
		try
			GetMem(MaskBuffer, SizeOf(COLORREF) * Width * SourceHeight);
			MaskDC := CreateCompatibleDC(ScreenDC);
			if MaskDC = 0 then
				RaiseLastOSError;

			Old := SelectObject(MaskDC, CursorInfo.hbmMask);
			GetDIBits(MaskDC, CursorInfo.hbmMask, 0, SourceHeight, MaskBuffer, CursorBitmapInfo, DIB_RGB_COLORS);
			SelectObject(MaskDC, Old);
			DeleteDC(MaskDC);

			if not IsMonochrome then
			begin
				GetMem(ColorBuffer, SizeOf(COLORREF) * Width * Height);

				ColorDC := CreateCompatibleDC(ScreenDC);
				if ScreenDC = 0 then
					RaiseLastOSError;

				Old := SelectObject(ColorDC, CursorInfo.hbmColor);
				GetDIBits(ColorDC, CursorInfo.hbmColor, 0, Height, ColorBuffer, CursorBitmapInfo, DIB_RGB_COLORS);
				SelectObject(ColorDC, Old);
				DeleteDC(ColorDC);
			end;
		finally
			ReleaseDC(0, ScreenDC);
		end;

		// Transfer cursor image into the surface
		Result.LockRect(Locked, nil, D3DLOCK_NOSYSLOCK);
		try
			p := Locked.pBits;
			for y := 0 to Height - 1 do
				for x := 0 to Width - 1 do
				begin
					if IsMonochrome then
					begin
						Color := MaskBuffer^[Width*(Height-1-y) + x];
						Mask := MaskBuffer^[Width*(SourceHeight-1-y) + x];
					end
					else
					begin
						Color := ColorBuffer^[Width*(Height-1-y) + x];
						Mask := MaskBuffer^[Width*(Height-1-y) + x];
					end;

					if Mask = 0 then
						p^[Width*y + x] := $ff000000 or Color
					else
						p^[Width*y + x] := 0;
				end
		finally
			Result.UnlockRect
		end;

		HR := Device.SetCursorProperties(CursorInfo.xHotspot, CursorInfo.yHotspot, Result);
		if FAILED(HR) then
			raise EOSError.Create(D3DXGetErrorString(HR));

	finally
		if MaskBuffer <> nil then FreeMem(MaskBuffer);
		if ColorBuffer <> nil then FreeMem(ColorBuffer);
		if CursorInfo.hbmMask <> 0 then DeleteObject(CursorInfo.hbmMask);
		if CursorInfo.hbmColor <> 0 then DeleteObject(CursorInfo.hbmColor);
	end
end;

function XYRectangle(const X1, Y1, X2, Y2: Single): TVertex4Array;
begin
	FillChar(Result, SizeOf(Result), 0);
	Result[0].x := X1;
	Result[0].y := Y1;
	Result[0].tv := 1;

	Result[1].x := X1;
	Result[1].y := Y2;

	Result[2].x := X2;
	Result[2].y := Y1;
	Result[2].tu := 1;
	Result[2].tv := 1;

	Result[3].x := X2;
	Result[3].y := Y2;
	Result[3].tu := 1;
end;

function YZRectangle(const Y1, Z1, Y2, Z2: Single; const X: Single = 0): TVertex4Array;
begin
	FillChar(Result, SizeOf(Result), 0);
	Result[0].x := X;
	Result[0].y := y1;
	Result[0].z := z1;
	Result[0].tv := 1;

	Result[1].x := X;
	Result[1].y := y1;
	Result[1].z := z2;

	Result[2].x := X;
	Result[2].y := y2;
	Result[2].z := z1;
	Result[2].tu := 1;
	Result[2].tv := 1;

	Result[3].x := X;
	Result[3].y := y2;
	Result[3].z := z2;
	Result[3].tu := 1;
end;

function ZXRectangle(const Z1, X1, Z2, X2: Single; const Y: Single = 0): TVertex4Array;
begin
	FillChar(Result, SizeOf(Result), 0);
	Result[0].x := X1;
	Result[0].y := y;
	Result[0].z := z1;
	Result[0].tv := 1;

	Result[1].x := X1;
	Result[1].y := y;
	Result[1].z := z2;

	Result[2].x := X2;
	Result[2].y := y;
	Result[2].z := z1;
	Result[2].tu := 1;
	Result[2].tv := 1;

	Result[3].x := X2;
	Result[3].y := y;
	Result[3].z := z2;
	Result[3].tu := 1;
end;

function Rectangle(const X1, Y1, X2, Y2: Integer; const U1, V1, U2, V2: Single): TVertex4Array;
begin
	//for 2D

	Result[0].x := X1;
	Result[0].y := Y2;
	Result[0].z := 0.5;
	Result[0].rhw := 1;
	Result[0].diffuse := $ffffffff;
	Result[0].specular := $ffffffff;
	Result[0].tu := U1;
	Result[0].tv := V2;

	Result[1].x := X1;
	Result[1].y := Y1;
	Result[1].z := 0.5;
	Result[1].rhw := 1;
	Result[1].diffuse := $ffffffff;
	Result[1].specular := $ffffffff;
	Result[1].tu := U1;
	Result[1].tv := V1;

	Result[2].x := X2;
	Result[2].y := Y2;
	Result[2].z := 0.5;
	Result[2].rhw := 1;
	Result[2].diffuse := $ffffffff;
	Result[2].specular := $ffffffff;
	Result[2].tu := U2;
	Result[2].tv := V2;

	Result[3].x := X2;
	Result[3].y := Y1;
	Result[3].z := 0.5;
	Result[3].rhw := 1;
	Result[3].diffuse := $ffffffff;
	Result[3].specular := $ffffffff;
	Result[3].tu := U2;
	Result[3].tv := V1;
end;

function ColorMaterial(const Ambient: TD3DColorValue): TD3DMaterial8;
begin
	FillChar(Result, SizeOf(Result), 0);
	Result.Ambient := Ambient;
	Result.Diffuse.a := 1;
end;

{ D3DX }

function LoadTexture(
	const Device: IDirect3DDevice8;
	const FileName: WideString;
	Format: TTextureFormat;
	TransparentColor: TD3DColor): ITexture;
var
	HR: HRESULT;
begin
	HR := D3DXCreateTextureFromFileExW(Device, PWideChar(FileName),
		0, 0, 0, 0, Format, D3DPOOL_MANAGED, D3DX_DEFAULT, D3DX_DEFAULT,
		TransparentColor, nil, nil, Result);

	if FAILED(HR) then
		raise EOSError.Create(D3DXGetErrorString(HR));
end;

function LoadTexture(
	const Device: IDirect3DDevice8;
	P: Pointer; Size: Cardinal;
	Format: TTextureFormat;
	TransparentColor: TD3DColor): ITexture;
var
	HR: HRESULT;
begin
	HR := D3DXCreateTextureFromFileInMemoryEx(Device, P^, Size,
		0, 0, 0, 0, Format, D3DPOOL_MANAGED, D3DX_DEFAULT, D3DX_DEFAULT,
		TransparentColor, nil, nil, Result);

	if FAILED(HR) then
		raise EOSError.Create(D3DXGetErrorString(HR));
end;

{ TLight }

procedure TLight.BeforeDestruction;
begin
	Exclude(TLights(Owner).FUsedID, ID);
	Exclude(TLights(Owner).FEnabled, ID);
	Modify;
	inherited;
end;

constructor TLight.Create(Screen: T3DScreen; AType: TLightType);
var
	I: Cardinal;
begin
	inherited Create(Screen.Lights, nil);

	FData._Type := AType;

	for I := 0 to 31 do
		if not(I in TLights(Owner).FUsedID) then
		begin
			FID := I;
			Break;
		end;

	Include(TLights(Owner).FUsedID, ID);
	Modify;
end;

function TLight.GetEnabled: Boolean;
begin
	Result := ID in TLights(Owner).FEnabled
end;

procedure TLight.Modify;
begin
	Include(TLights(Owner).FModified, ID)
end;

procedure TLight.SetAmbient(const Value: TD3DColorValue);
begin
	FData.Ambient := Value;
	Modify
end;

procedure TLight.SetDiffuse(const Value: TD3DColorValue);
begin
	FData.Diffuse := Value;
	Modify
end;

procedure TLight.SetDirection(const Value: TD3DVector);
begin
	FData.Direction := Value;
	Modify
end;

procedure TLight.SetEnabled(const Value: Boolean);
begin
	Include(TLights(Owner).FEnabled, ID);
	Modify
end;

procedure TLight.SetSpecular(const Value: TD3DColorValue);
begin
	FData.Specular := Value;
	Modify
end;

{ TLights }

procedure TLights.SetAmbient(const Value: TD3DCOLOR);
begin
	Assert(Owner is T3DScreen);
	T3DScreen(Owner).Direct3DDevice.SetRenderState(D3DRS_AMBIENT, Value)
end;

{ T3DScreen }

const
	ScreenClassName = 'DirectXGraphicsScreen';

function ScreenProc(Wnd: HWND; Msg: UINT; WP: WParam; LP: LParam): LRESULT; stdcall;

	function Self: T3DScreen;
	begin
		Result := T3DScreen(GetWindowLongW(Wnd, GWL_USERDATA))
	end;

begin
	case Msg of
	WM_CREATE:
	begin
		SetWindowLongW(Wnd, GWL_USERDATA, LongInt(PCREATESTRUCTW(LP)^.lpCreateParams));
		Result := 0;
	end;
	WM_ACTIVATE:
	begin
		with Self do
			if FullScreen and Standby then
				if WP = WA_ACTIVE then
					ShowWindow(Wnd, SW_MAXIMIZE)
				else
					ShowWindow(Wnd, SW_MINIMIZE);

		Result := DefWindowProcW(Wnd, Msg, WP, LP);
	end;
	WM_SIZE:
	begin
		with Self do
			if Standby then
			begin
				FinalizeDirectXGraphics;
				InitializeDirectXGraphics;
			end;

		Result := DefWindowProcW(Wnd, Msg, WP, LP);
	end;
	WM_SYSCHAR:
	begin
		if WP = 13 then
		begin
			with Self do FullScreen := not FullScreen;
			Result := 0;
		end
		else
		begin
			Result := DefWindowProcW(Wnd, Msg, WP, LP)
		end;
	end;
	WM_SETCURSOR:
	begin
		if not Self.Standby or (GetActiveWindow <> GetForegroundWindow) then
			Result := DefWindowProcW(Wnd, Msg, WP, LP)
		else
		begin
			SetCursor(0);
			with Self do Direct3DDevice.ShowCursor(CursorVisible);
			Result := 1;
		end;
	end;
	WM_MOUSEFIRST..WM_MOUSELAST:
	begin
		Result := Self.WindowProc(Msg, WP, LP);
	end;
	WM_SYSCOMMAND:
	begin
		if ((WP = SC_KEYMENU) or (WP = SC_MOUSEMENU) or (WP = SC_CLOSE)) and
			Self.Standby and Self.FullScreen then
			Result := 0
		else
			Result := DefWindowProcW(Wnd, Msg, WP, LP);
	end;
	else
		Result := DefWindowProcW(Wnd, Msg, WP, LP);
	end;
end;

procedure RegisterScreenWindow;
var
	WndClass: WNDCLASSEXW;
begin
	if not GetClassInfoExW(MainInstance, ScreenClassName, WndClass) then
	begin
		WndClass.cbSize := SizeOf(WndClass);
		WndClass.style := 0; //CS_SAVEBITS;
		WndClass.lpfnWndProc := @ScreenProc;
		WndClass.cbClsExtra := 0;
		WndClass.cbWndExtra := 0;
		WndClass.hInstance := MainInstance;
		WndClass.hIcon := LoadIconW(MainInstance, 'MAINICON');
		WndClass.hCursor := 0;
		WndClass.hbrBackground := GetStockObject(BLACK_BRUSH);
		WndClass.lpszMenuName := nil;
		WndClass.lpszClassName := ScreenClassName;
		WndClass.hIconSm := 0;
		RegisterClassExW(WndClass);
	end	
end;

procedure T3DScreen.BeginScene;
var
	P: TPoint;
begin
	if CursorVisible then
	begin
		GetCursorPos(P);
		Direct3DDevice.SetCursorPosition(P.X, P.Y, 0);
	end;
	Direct3DDevice.BeginScene
end;

procedure T3DScreen.Clear(Color: TD3DColor);
const
	Options: array[Boolean] of DWORD = (D3DCLEAR_TARGET, D3DCLEAR_TARGET or D3DCLEAR_ZBUFFER);
begin
	Direct3DDevice.Clear(0, nil, Options[FZBuffer], Color, 1, 0);
end;

procedure T3DScreen.ClearTexture;
begin
	Direct3DDevice.SetTexture(0, nil);
end;

procedure T3DScreen.ClearZBuffer;
begin
	Direct3DDevice.Clear(0, nil, D3DCLEAR_ZBUFFER, 0, 1, 0);
end;

constructor T3DScreen.Create(
	AParent: TAnyWindow;
	const APosition: TPoint;
	const ASize: TSize;
	const ACreator: IScreenCreator;
	AFullScreen: Boolean;
	const AFullScreenSize: TSize;
	AZBuffer: Boolean);
begin
	FFullScreen := AFullScreen;
	FZBuffer := AZBuffer;

	FFullScreenSize := AFullScreenSize;

	FCreator := ACreator;

	FLights := TLights.Create;
	FLights.Owner := Self;

	inherited Create(AParent, '', APosition, ASize);
end;

function T3DScreen.CreateTexture(W, H: Integer): IDirect3DTexture8;
begin
	Direct3DDevice.CreateTexture(W, H, 0, 0, D3DFMT_A1R5G5B5, D3DPOOL_SYSTEMMEM, Result)
end;

function T3DScreen.CreateVertexBuffer(const V: array of TVertex): IDirect3DVertexBuffer8;
var
	Size: Integer;
	P: PByte;
begin
	Size := SizeOf(TVertex) * Length(V);
	Direct3DDevice.CreateVertexBuffer(Size, 0, D3DFVF_VERTEX, D3DPOOL_MANAGED, Result);
	if Result.Lock(0, 0, P, D3DLOCK_NOSYSLOCK) = D3D_OK then
	begin
		//CopyMemory(P, @V[0], Size);
		Move(V[0], P^, Size);
		Result.Unlock;
	end
end;

procedure T3DScreen.Draw(const AVerteces: IDirect3DVertexBuffer8;
	AType: TPrimitiveType; Start, Count: Cardinal);
begin
	Direct3DDevice.SetStreamSource(0, AVerteces, SizeOf(TVertex));
	Direct3DDevice.DrawPrimitive(AType, Start, Count);
end;

procedure T3DScreen.Draw(const AVerteces: TVertex4Array);
begin
	Direct3DDevice.DrawPrimitiveUP(D3DPT_TRIANGLESTRIP, 2, AVerteces[0], SizeOf(TVertex));
end;

procedure T3DScreen.EndScene;
begin
	Flush;
	Direct3DDevice.EndScene
end;

procedure T3DScreen.Finalize;
begin
	FinalizeDirectXGraphics;
	DestroyWindow(FScreenWindow);
	inherited;
end;

procedure T3DScreen.FinalizeDirectXGraphics;
begin
	if not Standby then Exit;
	FStandby := False;

	if FCreator <> nil then
		FCreator.FinalizeScreen;

	FDirect3DDevice := nil;
	FDirect3D := nil;
end;

procedure T3DScreen.Flush;

	procedure Execute(Current: Integer);
	begin
		with FRenderingObjects[Current] do
		begin
			if Before >= 0 then Execute(Before);
			Direct3DDevice.SetTransform(D3DTS_WORLD, World);
			TRender(Render)(FRenderingObjects[Current]);
			if After >= 0 then Execute(After);
		end
	end;

begin
	if FRenderingObjects <> nil then
	begin
		Execute(0);
		FRenderingObjects := nil;
	end;
end;

function T3DScreen.GetBackBuffer: IDirect3DSurface8;
begin
	Direct3DDevice.GetBackBuffer(0, D3DBACKBUFFER_TYPE_MONO, Result)
end;

function T3DScreen.GetMaterial: TD3DMaterial8;
begin
  Direct3DDevice.GetMaterial(Result)
end;

function T3DScreen.GetProjectionMatrix: TD3DMatrix;
begin
	Direct3DDevice.GetTransform(D3DTS_PROJECTION, Result);
end;

function T3DScreen.GetRenderState(Index: TRenderState): Cardinal;
begin
	Direct3DDevice.GetRenderState(Index, Result)
end;

function T3DScreen.GetViewMatrix: TD3DMatrix;
begin
	Direct3DDevice.GetTransform(D3DTS_VIEW, Result);
end;

procedure T3DScreen.Initialize;
var
	S: DWORD;
	P: HWND;
begin
	inherited;

	S := WS_VISIBLE or WS_CHILD;
	P := Handle;
	if FullScreen then
	begin
		S := WS_POPUP or WS_SYSMENU;
		P := 0;
	end;

	RegisterScreenWindow;
	FScreenWindow := CreateWindowExW(WS_EX_TOPMOST or WS_EX_APPWINDOW,
						ScreenClassName, '', S,
						0, 0, 0, 0, P, 0, MainInstance, Self);

	PostMessageW(WM_LIGHTWINDOWS, LW_READY, 0);
end;

procedure T3DScreen.InitializeDirectXGraphics;
var
	DisplayMode: TD3DDisplayMode;
	PresentParameters: TD3DPresentParameters;
	I: Integer;
	Format: D3DFORMAT;
begin
	if Standby then Exit;

	FDirect3D := Direct3DCreate8(D3D_SDK_VERSION); //DX8.0 = 120

	if FullScreen then
	begin
		Format := D3DFMT_R3G3B2; {ň̃P[X}
		for I := Direct3D.GetAdapterModeCount(D3DADAPTER_DEFAULT) - 1 downto 0 do
		begin
			Direct3D.EnumAdapterModes(D3DADAPTER_DEFAULT, I, DisplayMode);
			if (DisplayMode.Format >= D3DFMT_R8G8B8) and (DisplayMode.Format < Format) then
				Format := DisplayMode.Format;
		end;

		FillChar(PresentParameters, SizeOf(PresentParameters), 0);
		PresentParameters.BackBufferWidth := ScreenSize.cx;
		PresentParameters.BackBufferHeight := ScreenSize.cy;
		PresentParameters.BackBufferFormat := Format;
		PresentParameters.BackBufferCount := 1;
		PresentParameters.SwapEffect := D3DSWAPEFFECT_DISCARD;
		PresentParameters.EnableAutoDepthStencil := FZBuffer;
		PresentParameters.AutoDepthStencilFormat := D3DFMT_D16;
		PresentParameters.FullScreen_RefreshRateInHz := DisplayMode.RefreshRate;
	end
	else
	begin
		Direct3D.GetAdapterDisplayMode(D3DADAPTER_DEFAULT, DisplayMode);
		FillChar(PresentParameters, SizeOf(PresentParameters), 0);
		PresentParameters.Windowed := True;
		PresentParameters.BackBufferFormat := DisplayMode.Format;
		PresentParameters.BackBufferCount := 1;
		PresentParameters.SwapEffect := D3DSWAPEFFECT_DISCARD;
		PresentParameters.EnableAutoDepthStencil := FZBuffer;
		PresentParameters.AutoDepthStencilFormat := D3DFMT_D16;
	end;

	if Direct3D.CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, FScreenWindow,
					D3DCREATE_SOFTWARE_VERTEXPROCESSING, PresentParameters, FDirect3DDevice) = D3D_OK then
		FStandby := True
	else if Direct3D.CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_REF, FScreenWindow,
					D3DCREATE_SOFTWARE_VERTEXPROCESSING, PresentParameters, FDirect3DDevice) = D3D_OK then
		FStandby := True;

	if Standby and (FCreator <> nil) then
	begin
		Direct3DDevice.SetVertexShader(D3DFVF_VERTEX);
		FCreator.InitializeScreen;
	end;
end;

function T3DScreen.OptimalProjectionMatrix: TD3DMatrix;
var
	R: Real;
	S: TSize;
begin
	S := ScreenSize;
	R := Sqrt(Sqr(S.CX) + Sqr(S.CY * 4 / 3)) / S.CX / Sqrt(2);
	R := ArcSin(R / Sqrt(2));
	if R < PI / 4 then R := PI / 4;
	if R > PI / 2 then R := PI / 2;
	//D3DXMatrixPerspectiveFovLH(Result, R, S.CY / S.CX, 1, 1000);
	LDXGraphics.SetProjectionMatrix(Result, R, S.CY / S.CX, 1, 1000);
end;

procedure T3DScreen.Present;
begin
	Direct3DDevice.Present(nil, nil, 0, nil);
end;

procedure T3DScreen.ReadyFog(ANear, AFar: Single; Color: TD3DColor);
begin
	Direct3DDevice.SetRenderState(D3DRS_FOGVERTEXMODE, D3DFOG_LINEAR);
	Direct3DDevice.SetRenderState(D3DRS_FOGSTART, PDWORD(@ANear)^);
	Direct3DDevice.SetRenderState(D3DRS_FOGEND, PDWORD(@AFar)^);
	Direct3DDevice.SetRenderState(D3DRS_FOGCOLOR, Color);
end;

procedure T3DScreen.ReadyLights;

	function Each(Obj: T0Object): Boolean cdecl;
	begin
		Assert(TObject(Obj) is TLight);

		if TLight(Obj).ID in Lights.FModified then
		begin
			Direct3DDevice.LightEnable(TLight(Obj).ID, TLight(Obj).ID in Lights.FEnabled);
			Direct3DDevice.SetLight(TLight(Obj).ID, TLight(Obj).FData);
			Exclude(Lights.FModified, TLight(Obj).ID);
		end;

		Result := False;
	end;

var
	Index: Integer;
begin
	if Lights.FModified <> [] then
	begin
		Lights.Enumerate(TLight, CEachObject(@Each), GetContext);

		for Index := 0 to 31 do
			if Index in Lights.FModified then
				Direct3DDevice.LightEnable(Index, False);

		Lights.FModified := [];
	end;
end;

procedure T3DScreen.ReadyStatesForAdd;
begin
	with Direct3DDevice do
	begin
		SetTextureStageState(0,D3DTSS_COLOROP,D3DTOP_MODULATE);
		SetTextureStageState(0,D3DTSS_ALPHAARG2,D3DTA_CURRENT);
		SetTextureStageState(0,D3DTSS_ALPHAOP,D3DTOP_SELECTARG1);
		SetRenderState(D3DRS_ALPHABLENDENABLE,1);
		SetRenderState(D3DRS_SRCBLEND,D3DBLEND_ONE);
		SetRenderState(D3DRS_DESTBLEND,D3DBLEND_ONE);
	end
end;

procedure T3DScreen.ReadyStatesForAlpha;
var
	Mt: TD3DMaterial8;
begin
	with Direct3DDevice do
	begin
		SetRenderState(D3DRS_ALPHATESTENABLE, 1);
		SetRenderState(D3DRS_ALPHAREF, 0);
		SetRenderState(D3DRS_ALPHAFUNC, D3DCMP_GREATER);
		SetRenderState(D3DRS_ALPHABLENDENABLE, 1);
		SetRenderState(D3DRS_SRCBLEND, D3DBLEND_SRCALPHA);
		SetRenderState(D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA);

		SetTextureStageState(0, D3DTSS_ALPHAOP, D3DTOP_BLENDDIFFUSEALPHA);
		SetTextureStageState(0, D3DTSS_ALPHAARG1, D3DTA_TEXTURE);
		SetTextureStageState(0, D3DTSS_ALPHAARG2, D3DTA_CURRENT);

		FillChar(Mt, SizeOf(Mt), 0);
		Mt.Ambient.r := 1; Mt.Ambient.g := 1; Mt.Ambient.b := 1;
		Mt.Diffuse.a := 1;
		SetMaterial(Mt);
	end
end;

procedure T3DScreen.ReadyStatesForMultiply;
begin
	with Direct3DDevice do
	begin
		SetTextureStageState(0,D3DTSS_COLOROP,D3DTOP_MODULATE);
		SetTextureStageState(0,D3DTSS_ALPHAARG2,D3DTA_CURRENT);
		SetTextureStageState(0,D3DTSS_ALPHAOP,D3DTOP_SELECTARG1);
		SetRenderState(D3DRS_ALPHABLENDENABLE,1);
		SetRenderState(D3DRS_SRCBLEND,D3DBLEND_ZERO);
		SetRenderState(D3DRS_DESTBLEND,D3DBLEND_SRCCOLOR);
	end
end;

procedure T3DScreen.ReadyStatesForNormal;
begin
	with Direct3DDevice do
	begin
		SetRenderState(D3DRS_ALPHATESTENABLE, 0);
		SetRenderState(D3DRS_ALPHABLENDENABLE, 0);

		SetTextureStageState(0, D3DTSS_COLOROP, D3DTOP_MODULATE);
		SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
		SetTextureStageState(0, D3DTSS_COLORARG2, D3DTA_DIFFUSE);
		SetTextureStageState(0, D3DTSS_ALPHAOP, D3DTOP_DISABLE);
	end
end;

procedure T3DScreen.ReadyStatesForSubtract;
begin
	with Direct3DDevice do
	begin
		SetTextureStageState(0,D3DTSS_COLOROP,D3DTOP_MODULATE);
		SetTextureStageState(0,D3DTSS_ALPHAARG2,D3DTA_CURRENT);
		SetTextureStageState(0,D3DTSS_ALPHAOP,D3DTOP_SELECTARG1);
		SetRenderState(D3DRS_ALPHABLENDENABLE,1);
		SetRenderState(D3DRS_SRCBLEND,D3DBLEND_ZERO);
		SetRenderState(D3DRS_DESTBLEND,D3DBLEND_INVSRCCOLOR);
	end
end;

procedure T3DScreen.Realign;
var
	Size: TSize;
begin
	Size := ClientSize;
	MoveWindow(FScreenWindow, 0, 0, Size.cx, Size.cy, False);
end;

procedure T3DScreen.Render(const ARender: TRender; const AWorld: TD3DMatrix; ABlending: Boolean; AData: Pointer);
var
	Work: TD3DVector;
	NewIndex, Current: Integer;
	Over: Boolean;
begin
	Work.x := AWorld._41; Work.y := AWorld._42; Work.z := AWorld._43;
	Work := VectorMulMatrix(Work, ViewMatrix);

	if Work.z >= 0 then
	begin
		NewIndex := Length(FRenderingObjects);
		SetLength(FRenderingObjects, NewIndex + 1);
		with FRenderingObjects[NewIndex] do
		begin
			Render := TMethod(ARender);
			World := AWorld;
			Screen := Self;
			z := Work.z;
			absx := Abs(Work.x);
			Blending := ABlending;
			Data := AData;
			Before := -1;
			After := -1;
		end;
		if NewIndex > 0 then
		begin
			Current := 0;
			repeat
				if ABlending then
				begin
					Over := not FRenderingObjects[Current].Blending;
					if not Over then
					begin
						Over := Work.z < FRenderingObjects[Current].z;
						if not Over then
							Over := (Work.z = FRenderingObjects[Current].z) and
											(FRenderingObjects[NewIndex].absx < FRenderingObjects[Current].absx);
					end
				end
				else
					Over := not FRenderingObjects[Current].Blending and (Work.z > FRenderingObjects[Current].z);

				if not Over then
				begin
					if FRenderingObjects[Current].Before < 0 then
					begin
						FRenderingObjects[Current].Before := NewIndex;
						Break;
					end;
					Current := FRenderingObjects[Current].Before
				end
				else
				begin
					if FRenderingObjects[Current].After < 0 then
					begin
						FRenderingObjects[Current].After := NewIndex;
						Break;
					end;
					Current := FRenderingObjects[Current].After;
				end
			until False
		end;
	end
end;

function T3DScreen.ScreenCursorPos: TPoint;
var
	R: TRect;
begin
	GetCursorPos(Result);
	if not FullScreen then
	begin
		GetWindowRect(Handle, R);
		Dec(Result.X, R.Left);
		Dec(Result.Y, R.Top)
	end;
end;

function T3DScreen.ScreenSize: TSize;
var
	R: TRect;
begin
	if FullScreen then
		Result := FFullScreenSize
	else
	begin
		GetWindowRect(FScreenWindow, R);
		Result.cx := R.Right - R.Left;
		Result.cy := R.Bottom - R.Top
	end
end;

procedure T3DScreen.SetFog(const Value: Boolean);
begin
	Direct3DDevice.SetRenderState(D3DRS_FOGENABLE, Ord(Value));
end;

procedure T3DScreen.SetFullScreen(const Value: Boolean);

	procedure ReadyFullScreen;
	var
		Info: MENUITEMINFOW;
	begin
		ShowWindow(FScreenWindow, SW_HIDE);
		Windows.SetParent(FScreenWindow, 0);
		SetWindowLongW(FScreenWindow, GWL_STYLE, Integer(WS_POPUP or WS_SYSMENU));

		Info.cbSize := SizeOf(Info);
		Info.fMask := MIIM_STATE;
		Info.fState := MFS_DISABLED;
		SetMenuItemInfoW(GetSystemMenu(FScreenWindow, False), SC_CLOSE, False, Info);

		ShowWindow(FScreenWindow, SW_MAXIMIZE);
		TopParent.Hide;
	end;

begin
	if Value = FFullScreen then Exit;

	FinalizeDirectXGraphics;
	FFullScreen := Value;

	if FFullScreen then
	begin
		ReadyFullScreen
	end
	else
	begin
		ShowWindow(FScreenWindow, SW_HIDE);
		SetWindowLongW(FScreenWindow, GWL_STYLE, WS_CHILD);
		Windows.SetParent(FScreenWindow, Handle);
		Realign;
		ShowWindow(FScreenWindow, SW_SHOW);
		ShowWindow(TopParent.Handle, SW_SHOW);
	end;

	InitializeDirectXGraphics;
end;

procedure T3DScreen.SetMaterial(const Value: TD3DMaterial8);
begin
	Direct3DDevice.SetMaterial(Value)
end;

procedure T3DScreen.SetProjectionMatrix(const Value: TD3DMatrix);
begin
	Direct3DDevice.SetTransform(D3DTS_PROJECTION, Value);
end;

procedure T3DScreen.SetRenderState(Index: TRenderState; const Value: Cardinal);
begin
	Direct3DDevice.SetRenderState(Index, Value)
end;

procedure T3DScreen.SetTexture(const Value: IDirect3DTexture8);
begin
	Direct3DDevice.SetTexture(0, Value)
end;

procedure T3DScreen.SetTextures(Index: Integer; const Value: IDirect3DTexture8);
begin
	Direct3DDevice.SetTexture(Index, Value)
end;

procedure T3DScreen.SetViewMatrix(const Value: TD3DMatrix);
begin
	Direct3DDevice.SetTransform(D3DTS_VIEW, Value);
end;

procedure T3DScreen.WMLightWindows(var Message: TMessage);
begin
	if Message.WParam = LW_READY then
	begin
		if FullScreen then
		begin
			TopParent.Hide;
			ShowWindow(ScreenWindow, SW_MAXIMIZE);
		end;
		InitializeDirectXGraphics;
	end
	else
		inherited
end;

procedure T3DScreen.WMSetFocus(var Message);
begin
	Windows.SetFocus(FScreenWindow);
end;

end.
