unit LDXInput;

interface

uses
	DirectInput, SysUtils, ActiveX,
	SysHacks,
	LObjects, LWindows;

type
	TInputState = record
		X, Y, Z: Integer;
		Buttons: array[0..2] of record
			Pushed: Boolean;
			Event: (beContinued, bePushed, beReleased);
		end;
	end;

	TMouseInput = class;
	TKeyboardInput = class;

	TInputManager = class(T0Collection)
	private
		FDirectInput: IDirectInput8W;
		FSystemKeyboard: TKeyboardInput;
		FSystemMouse: TMouseInput;
		FIsReady: Boolean;
	public
		procedure AfterConstruction; override;
		destructor Destroy; override;
		function SystemMouse: TMouseInput;
		function SystemKeyboard: TKeyboardInput;
		procedure UpdateState(var State: TInputState);
		property DirectInput: IDirectInput8W read FDirectInput;
		property IsReady: Boolean read FIsReady;
	end;

	T0InputDevice = class(T0CollectionItem)
	private
		FDirectInputDevice: IDirectInputDevice8W;
		FAcquiring: Boolean;
	protected
		procedure AddState(var State: TInputState); virtual; abstract;
	public
		constructor Create(AManager: TInputManager; const DeviceID: TGUID);
		procedure BeforeDestruction; override;
		procedure Acquire;
		procedure Leave;
		procedure Update; virtual; abstract;
		property DirectInputDevice: IDirectInputDevice8W read FDirectInputDevice;
		property Acquiring: Boolean read FAcquiring;
	end;

	TMouseInput = class(T0InputDevice)
	private
		FState: TDIMouseState2;
		function GetButtons(AIndex: Integer): Boolean;
	protected
		procedure AddState(var State: TInputState); override;
	public
		procedure AfterConstruction; override;
		procedure Update; override;
		property State: TDIMouseState2 read FState;
		property X: Integer read FState.lx;
		property Y: Integer read FState.ly;
		property Z: Integer read FState.lz;
		property Buttons[AIndex: Integer]: Boolean read GetButtons;
		property LButton: Boolean index 0 read GetButtons;
		property RButton: Boolean index 1 read GetButtons;
		property MButton: Boolean index 2 read GetButtons;
	end;

	TDIKeyboardState = array[0..255] of Byte;

	TKeyboardInput = class(T0InputDevice)
	private
		FState: TDIKeyboardState;
		function GetKeys(AIndex: Byte): Boolean;
	protected
		procedure AddState(var State: TInputState); override;
	public
		procedure AfterConstruction; override;
		procedure Update; override;
		property State: TDIKeyboardState read FState;
		property Keys[AIndex: Byte]: Boolean read GetKeys;
	end;

implementation

{ TInputManager }

procedure TInputManager.AfterConstruction;
begin
	inherited;
	FIsReady := SUCCEEDED(DirectInput8Create(
		MainInstance, DIRECTINPUT_VERSION, IID_IDirectInput8W, FDirectInput, nil));
end;

destructor TInputManager.Destroy;
begin
	Clear;
	FDirectInput := nil;
	inherited;
end;

function TInputManager.SystemKeyboard: TKeyboardInput;
begin
	if FSystemKeyboard = nil then
		FSystemKeyboard := TKeyboardInput.Create(Self, GUID_SysKeyboard);

	Result := FSystemKeyboard
end;

function TInputManager.SystemMouse: TMouseInput;
begin
	if FSystemMouse = nil then
		FSystemMouse := TMouseInput.Create(Self, GUID_SysMouse);

	Result := FSystemMouse
end;

procedure TInputManager.UpdateState(var State: TInputState);

	function EnumProc(Item: T0CollectionItem): Boolean; cdecl;
	begin
		Assert(Item is T0InputDevice);

		if T0InputDevice(Item).Acquiring then
		begin
			T0InputDevice(Item).Update;
			T0InputDevice(Item).AddState(State);
		end;

		Result := False;
	end;

var
	OldState: TInputState;
	I: Integer;
begin
	OldState := State;
	State.Buttons[0].Pushed := False;
	State.Buttons[1].Pushed := False;
	State.Buttons[2].Pushed := False;
	Enumerate(T0InputDevice, CEachObject(@EnumProc), GetContext);
	for I := 0 to 2 do
		if OldState.Buttons[I].Pushed and not State.Buttons[I].Pushed then
			State.Buttons[I].Event := beReleased
		else if not OldState.Buttons[I].Pushed and State.Buttons[I].Pushed then
			State.Buttons[I].Event := bePushed
		else
			State.Buttons[I].Event := beContinued;
end;

{ T0InputDevice }

procedure T0InputDevice.Acquire;
begin
	DirectInputDevice.Acquire;
	FAcquiring := True
end;

procedure T0InputDevice.BeforeDestruction;
begin
	FDirectInputDevice.Unacquire;
	FDirectInputDevice := nil;
	inherited;
end;

constructor T0InputDevice.Create(AManager: TInputManager;	const DeviceID: TGUID);
begin
	inherited Create(AManager);
	AManager.DirectInput.CreateDevice(DeviceID, FDirectInputDevice, nil);
end;

procedure T0InputDevice.Leave;
begin
	DirectInputDevice.Unacquire;
	FAcquiring := False
end;

{ TMouseInput }

procedure TMouseInput.AddState(var State: TInputState);
begin
	Inc(State.X, X);
	Inc(State.Y, Y);
	Inc(State.Z, Z);
	State.Buttons[0].Pushed := State.Buttons[0].Pushed or Buttons[0];
	State.Buttons[1].Pushed := State.Buttons[1].Pushed or Buttons[1];
	State.Buttons[2].Pushed := State.Buttons[2].Pushed or Buttons[2];
end;

procedure TMouseInput.AfterConstruction;
begin
	inherited;

	DirectInputDevice.SetCooperativeLevel(
					(CurrentThread as T0GUIThread).MainWindow.Handle,
					DISCL_BACKGROUND or DISCL_NONEXCLUSIVE);

	DirectInputDevice.SetDataFormat(c_dfDIMouse2);
end;

function TMouseInput.GetButtons(AIndex: Integer): Boolean;
begin
	Result := FState.rgbButtons[AIndex] and $80 <> 0
end;

procedure TMouseInput.Update;
begin
	DirectInputDevice.GetDeviceState(SizeOf(FState), @FState)
end;

{ TKeyboardInput }

procedure TKeyboardInput.AddState(var State: TInputState);
begin
	if Keys[DIK_LEFT ] then Inc(State.X, 100);
	if Keys[DIK_RIGHT] then Dec(State.X, 100);

	if Keys[DIK_UP   ] then Inc(State.Y, 100);
	if Keys[DIK_DOWN ] then Dec(State.Y, 100);

	if Keys[DIK_PGUP ] then Inc(State.Z, 100);
	if Keys[DIK_PGDN ] then Dec(State.Z, 100);

	State.Buttons[0].Pushed := State.Buttons[0].Pushed or Keys[DIK_Z];
	State.Buttons[1].Pushed := State.Buttons[1].Pushed or Keys[DIK_X];
	State.Buttons[2].Pushed := State.Buttons[2].Pushed or Keys[DIK_C];
end;

procedure TKeyboardInput.AfterConstruction;
begin
	inherited;

	DirectInputDevice.SetCooperativeLevel(
					(CurrentThread as T0GUIThread).MainWindow.Handle,
 					DISCL_BACKGROUND or DISCL_NONEXCLUSIVE);

	DirectInputDevice.SetDataFormat(c_dfDIKeyboard);
end;

function TKeyboardInput.GetKeys(AIndex: Byte): Boolean;
begin
	Result := FState[AIndex] and $80 <> 0
end;

procedure TKeyboardInput.Update;
begin
	DirectInputDevice.GetDeviceState(SizeOf(FState), @FState)
end;

end.
