unit LGraphics;

{*******************************************************}
{                                                       }
{       LightWindows                                    }
{       GDI Graphics                                    }
{                                                       }
{       1998..2004 YT                                   }
{                                                       }
{*******************************************************}

interface

uses
	Windows, SysUtils,
	AlgUtils,
	LObjects;

type
	TColor = record
		Value: COLORREF;
		SysColor: Integer;
	end;

{
	RGB          : $00RRGGBB
	PaletteIndex : $010000II
	PaletteRGB   : $02BBGGRR
}

function RGBColor(Value: COLORREF): TColor; inline;
function SystemColor(Kind: Cardinal): TColor; inline;

type
	PGDISharedHandle = ^TGDISharedHandle;
	TGDISharedHandle = record
		Next: PGDISharedHandle;
		RefCount: Integer;
		Handle: THandle;
		HashCode: THashCode;
	end;

	T0GDIObject = class(T0Object)
	private
		FHandle: THandle;
		FSharedHandle: PGDISharedHandle;
		function GetHandle: THandle;
	protected
		procedure Change;
		procedure Allocate; virtual;
		procedure Release; virtual; 
	public
		procedure BeforeDestruction; override;
		property Handle: THandle read GetHandle;
	end;

type
	TAnyFont = class(T0GDIObject)
	private
		FLog: LOGFONTW;
		function GetFaceName: WideString;
		function GetSize: Integer;
		function GetPitch: Byte;
	public
		constructor Create(AHandle: HFONT);
		property FaceName: WideString read GetFaceName;
		property CharSet: Byte read FLog.lfCharSet;
		property Height: Integer read FLog.lfHeight;
		property Size: Integer read GetSize;
		property Pitch: Byte read GetPitch;
		property Log: LOGFONTW read FLog;
		class function PixelsPerInch: Integer;
	end;

	TFont = class(TAnyFont)
	private
		procedure SetSize(ASize: Integer);
		procedure SetHeight(AHeight: Integer);
		procedure SetCharSet(ACharSet: Byte);
		procedure SetFaceName(const AName: WideString);
		procedure SetPitch(const Value: Byte);
	protected
		procedure Allocate; override;
		procedure Release; override;
	public
		constructor Create;
		procedure Assign(Source: HFONT); overload;
		procedure Assign(Source: TAnyFont); overload;
		procedure Assign(const Source: LOGFONTW); overload;
		property FaceName write SetFaceName;
		property CharSet write SetCharSet;
		property Height write SetHeight;
		property Size write SetSize;
		property Pitch write SetPitch;
	end;

type
	TAnyBrush = class(T0GDIObject)
	private
		FLog: LOGBRUSH;
	public
		constructor Create(AHandle: HBRUSH);
		property Color: COLORREF read FLog.lbColor;
		property Style: UINT read FLog.lbStyle;
		property Log: LOGBRUSH read FLog;
	end;

	TBrush = class(TAnyBrush)
	private
		procedure SetColor(Value: COLORREF);
		procedure SetStyle(Value: UINT);
	protected
		procedure Allocate; override;
		procedure Release; override;
	public
		constructor Create;
		procedure Assign(Source: HBRUSH); overload;
		procedure Assign(Source: TAnyBrush); overload;
		procedure Assign(const Source: LOGBRUSH); overload;
		property Color write SetColor;
		property Style write SetStyle;
	end;

type
	TAnyPen = class(T0GDIObject)
	private
		FLog: LOGPEN;
	public
		constructor Create(AHandle: HPEN);
		property Color: COLORREF read FLog.lopnColor;
		property Log: LOGPEN read FLog;
	end;

	TPen = class(TAnyPen)
	private
		procedure SetColor(Value: COLORREF);
	protected
		procedure Allocate; override;
		procedure Release; override;
	public
		constructor Create;
		procedure Assign(Source: HPEN); overload;
		procedure Assign(Source: TAnyPen); overload;
		procedure Assign(const Source: LOGPEN); overload;
		property Color write SetColor;
	end;

implementation

threadvar
	FontList: PGDISharedHandle;
	BrushList: PGDISharedHandle;
	PenList: PGDISharedHandle;

var
	_PixelsPerInch: Integer = 0;

function RGBColor(Value: COLORREF): TColor;
begin
	Result.Value    := Value;
	Result.SysColor := -1;
end;

function SystemColor(Kind: Cardinal): TColor;
begin
	Result.Value := GetSysColor(Kind);
	Result.SysColor := Kind;
end;

function AllocateGDIHandle(var List: PGDISharedHandle; const Data; Size: Integer): PGDISharedHandle;
var
	LogDataBuffer: record
	case Integer of
	0: (Font: LOGFONTW);
	1: (Brush: LOGBRUSH);
	2: (Pen: LOGPEN);
	end;

	Hash: THashCode;
	I: PGDISharedHandle;
begin
	Hash := GetHashCode(Data, Size);

	I := List;
	while I <> nil do
	begin
		if Hash = I.HashCode then
		begin
			GetObjectW(I.Handle, Size, @LogDataBuffer);
			if CompareMem(@Data, @LogDataBuffer, Size) then
			begin
				Result := I;
				Inc(Result.RefCount);
				Exit;
			end;
		end;
		I := I.Next;
	end;

	New(Result);
	Result.Next := List;
	Result.RefCount := 1;
	Result.Handle := 0; {I}
	Result.HashCode := Hash;

	List := Result;
end;

procedure ReleaseGDIHandle(var List: PGDISharedHandle; Handle: PGDISharedHandle);
var
	I: PGDISharedHandle;
begin
	Dec(Handle.RefCount);

	if Handle.RefCount = 0 then
	begin
		if List = Handle then
			List := Handle.Next
		else
		begin
			I := List;
			repeat
				if I^.Next = Handle then
				begin
					I^.Next := Handle^.Next;
					Break;
				end;
				I := I^.Next;
			until False;
		end;

		DeleteObject(Handle.Handle);
		Dispose(Handle);
	end
end;

{ T0GDIObject }

procedure T0GDIObject.Allocate;
begin
end;

procedure T0GDIObject.BeforeDestruction; {override}
begin
	if FSharedHandle <> nil then
		Release;

	inherited;
end;

procedure T0GDIObject.Change;
begin
	if FSharedHandle <> nil then
	begin
		Release;
		FSharedHandle := nil;
		FHandle := 0;
	end;
end;

function T0GDIObject.GetHandle: THandle;
begin
	if FHandle = 0 then
	begin
		Allocate;
		FHandle := FSharedHandle.Handle;
	end;
	Result := FHandle;
end;

procedure T0GDIObject.Release;
begin
end;

{ TAnyFont }

constructor TAnyFont.Create(AHandle: HFONT);
begin
	inherited Create;
	if AHandle <> 0 then
	begin
		FHandle := AHandle;
		GetObjectW(AHandle, SizeOf(FLog), @FLog);
	end
end;

function TAnyFont.GetFaceName: WideString;
begin
	Result:= FLog.lfFaceName;
end;

function TAnyFont.GetPitch: Byte;
begin
	Result := FLog.lfPitchAndFamily and 3;
end;

function TAnyFont.GetSize: Integer;
begin
	Result := - MulDiv(FLog.lfHeight, 72, PixelsPerInch)
end;

class function TAnyFont.PixelsPerInch: Integer;
var
	DC: HDC;
begin
	if _PixelsPerInch = 0 then
	begin
		{ĂȂAՓ˂Đݒ肷邱ƂɂȂĂ薳낤}
		DC := GetDC(0);
		_PixelsPerInch := GetDeviceCaps(DC, LOGPIXELSY);
		ReleaseDC(0, DC);
	end;
	Result := _PixelsPerInch;
end;

{ TFont }

procedure TFont.Allocate;
begin
	FSharedHandle := AllocateGDIHandle(FontList, FLog, SizeOf(FLog));
	if FSharedHandle.Handle = 0 then
		FSharedHandle.Handle := CreateFontIndirectW(FLog);
end;

procedure TFont.Assign(Source: HFONT);
begin
	Change;

	GetObjectW(Source, SizeOf(FLog), @FLog);
end;

procedure TFont.Assign(Source: TAnyFont);
begin
	Change;

	FLog := Source.Log;
	if Source.FSharedHandle <> nil then
	begin
		FHandle := Source.FHandle;
		FSharedHandle := Source.FSharedHandle;
		Inc(FSharedHandle.RefCount);
	end;
end;

procedure TFont.Assign(const Source: LOGFONTW);
begin
	Change;

	FLog := Source;
end;

constructor TFont.Create;
begin
	inherited Create(0)
end;

procedure TFont.Release;
begin
	ReleaseGDIHandle(FontList, FSharedHandle);
end;

procedure TFont.SetCharSet(ACharSet: Byte);
begin
	Change;

	FLog.lfCharSet:= ACharSet;
end;

procedure TFont.SetFaceName(const AName: WideString);
begin
	Change;

	FillChar(FLog.lfFaceName, SizeOf(FLog.lfFaceName), 0);
	{ĂȂ #0 ȍ~ɃS~cꍇ Hash Code ςĂ܂}

	lstrcpynW(FLog.lfFaceName, PWideChar(AName), High(FLog.lfFaceName) - 1);
end;

procedure TFont.SetHeight(AHeight: Integer);
begin
	Change;

	FLog.lfHeight:= AHeight;
end;

procedure TFont.SetPitch(const Value: Byte);
begin
	Change;

	Assert(Value and not 3 = 0);
	FLog.lfPitchAndFamily := FLog.lfPitchAndFamily and not 3 or Value;
end;

procedure TFont.SetSize(ASize: Integer);
begin
	Change;

	FLog.lfHeight := - MulDiv(ASize, PixelsPerInch, 72);
end;

{ TAnyBrush }

constructor TAnyBrush.Create(AHandle: HBRUSH);
begin
	inherited Create;
	if AHandle <> 0 then
	begin
		FHandle := AHandle;
		GetObjectW(AHandle, SizeOf(FLog), @FLog);
	end
end;

{ TBrush }

procedure TBrush.Allocate;
begin
	FSharedHandle := AllocateGDIHandle(BrushList, FLog, SizeOf(FLog));
	if FSharedHandle.Handle = 0 then
		FSharedHandle.Handle := CreateBrushIndirect(FLog);
end;

procedure TBrush.Assign(Source: HBRUSH);
begin
	Change;

	GetObjectW(Source, SizeOf(LOGBRUSH), @FLog);
end;

procedure TBrush.Assign(Source: TAnyBrush);
begin
	Change;

	FLog := Source.Log;
	if Source.FSharedHandle <> nil then
	begin
		FHandle := Source.FHandle;
		FSharedHandle := Source.FSharedHandle;
		Inc(FSharedHandle.RefCount);
	end;
end;

procedure TBrush.Assign(const Source: LOGBRUSH);
begin
	Change;

	FLog := Source
end;

constructor TBrush.Create;
begin
	inherited Create(0)
end;

procedure TBrush.Release;
begin
	ReleaseGDIHandle(BrushList, FSharedHandle)
end;

procedure TBrush.SetColor(Value: COLORREF);
begin
	if FLog.lbColor <> Value then
	begin
		Change;
		FLog.lbColor := Value
	end
end;

procedure TBrush.SetStyle(Value: UINT);
begin
	if FLog.lbStyle <> Value then
	begin
		Change;
		FLog.lbStyle := Value
	end
end;

{ TAnyPen }

constructor TAnyPen.Create(AHandle: HPEN);
begin
	inherited Create;
	if AHandle <> 0 then
	begin
		FHandle := AHandle;
		GetObjectW(AHandle, SizeOf(FLog), @FLog);
	end
end;

{ TPen }

procedure TPen.Allocate;
begin
	FSharedHandle := AllocateGDIHandle(PenList, FLog, SizeOf(FLog));
	if FSharedHandle.Handle = 0 then
		FSharedHandle.Handle := CreatePenIndirect(FLog)
end;

procedure TPen.Assign(Source: HPEN);
begin
	Change;

	GetObjectW(Source, SizeOf(FLog), @FLog)
end;

procedure TPen.Assign(Source: TAnyPen);
begin
	Change;

	FLog := Source.Log;
	if Source.FSharedHandle <> nil then
	begin
		FHandle := Source.FHandle;
		FSharedHandle := Source.FSharedHandle;
		Inc(FSharedHandle.RefCount);
	end;
end;

procedure TPen.Assign(const Source: LOGPEN);
begin
	Change;

	FLog := Source;
end;

constructor TPen.Create;
begin
	inherited Create(0)
end;

procedure TPen.Release;
begin
	ReleaseGDIHandle(PenList, FSharedHandle)
end;

procedure TPen.SetColor(Value: COLORREF);
begin
	Change;

	FLog.lopnColor := Value;
end;

end.

