unit LObjects;

{*******************************************************}
{                                                       }
{       Light Windows Core                              }
{       Base Objects                                    }
{                                                       }
{       1998..2006 YT                                   }
{                                                       }
{*******************************************************}

{$DEBUGINFO OFF}
{$LOCALSYMBOLS OFF}

interface

uses
	{$IFOPT C+} Windows, {$ENDIF}
	SysUtils, Types, SysConst,
	AlgUtils,
	LConsts;

{$INCLUDE LInterfaces.inc}

type
	T0Object = class;

	CEachObject = function(Obj: T0Object; Context: Pointer): Boolean cdecl;

	T0Object = class(TObject)
	private
		FOwner: T0Object;
		FFirst: T0Object;
		FLast: T0Object;
		FPred: T0Object;
		FSucc: T0Object;
{$IFOPT C+}
		FRefCount: Integer;
{$ENDIF}
		procedure SetOwner(Value: T0Object);
	protected
		function QueryInterface(const IID: TGUID; out Obj): HResult; stdcall;
		function _AddRef: Integer; {$IFOPT C+} virtual; {$ENDIF} stdcall;
		function _Release: Integer; {$IFOPT C+} virtual; {$ENDIF} stdcall;
		procedure FreeObjects;
		property LastObject: T0Object read FLast;
{$IFOPT C+}
		property _RefCount: Integer read FRefCount;
{$ENDIF}
	public
		destructor Destroy; override;
		function EnumerateObjects(Proc: CEachObject; Context: Pointer): T0Object;
		function HasObjects: Boolean;
		property Owner: T0Object read FOwner write SetOwner;
	end;

{RNV}
type
	T0Collection = class;

	T0CollectionItem = class(T0Object)
	public
		constructor Create(ACollection: T0Collection; ANext: T0CollectionItem = nil);
		destructor Destroy; override;
	end;

	T0Collection = class(T0Object)
	private
		FCount: Integer;
	protected
		function Enumerate(ItemClass: TClass; Proc: CEachObject; Context: Pointer): T0Object;
	public
		procedure Clear;
		property Count: Integer read FCount;
	end;

{ T0AggregatedObject }

type
	T0AggregatedObject = class(T0Object, IInterface)
	private
		FController: Pointer;
		procedure SetController(const Value: IInterface);
	protected
		function QueryInterface(const IID: TGUID; out Obj): HResult; stdcall;
		function _AddRef: Integer; {$IFOPT C+} override; {$ENDIF} stdcall;
		function _Release: Integer; {$IFOPT C+} override; {$ENDIF} stdcall;
	public
		property Controller: IInterface write SetController;
	end;

{ TObservable }

type
	TObservable = class(T0AggregatedObject, IObservable)
	private
		FViews: array of IObserver;
	protected
		procedure ViewIncreased; virtual;
		procedure ViewDecreased; virtual;
	public
		procedure Attach(const AView: IObserver);
		procedure Detach(const AView: IObserver);
		procedure Notify;
		function HasClients: Boolean;
	end;

var
	Application: IInterface;

threadvar
	CurrentThread: IInterface;

implementation

{ T0Object }

destructor T0Object.Destroy; {override}
begin
{$IFOPT C+}
	Assert(FRefCount = 0, Format(SInvalidRefCount, [ClassName, FRefCount]));
{$ENDIF}

	//Zƕʂ
	if FPred <> nil then FPred.FSucc:= FSucc;
	if FSucc <> nil then FSucc.FPred:= FPred;

	//eƉ؂
	if FOwner <> nil then
	begin
		if FOwner.FFirst = Self then FOwner.FFirst:= FSucc;
		if FOwner.FLast = Self then FOwner.FLast:= FPred;
	end;

	//qEĎ
	FreeObjects;

	inherited
end;

procedure T0Object.SetOwner(Value: T0Object);
begin
	Assert(FOwner = nil);

	FOwner := Value;

	if FOwner.FLast = nil then
		FOwner.FFirst := Self
	else
	begin
		FPred := FOwner.FLast;
		FPred.FSucc := Self;
	end;
	FOwner.FLast := Self;
end;

function T0Object.QueryInterface(const IID: TGUID; out Obj): HResult;
begin
	if GetInterface(IID, Obj) then Result := 0 else Result := E_NOINTERFACE;
end;

function T0Object._AddRef: Integer;
begin
{$IFOPT C+}
	Result := InterlockedIncrement(FRefCount);
{$ELSE}
	Result := 0
{$ENDIF}
end;

function T0Object._Release: Integer;
begin
{$IFOPT C+}
  Result := InterlockedDecrement(FRefCount);
{$ELSE}
	Result := 0
{$ENDIF}
end;

function T0Object.HasObjects: Boolean;
begin
	Result := FFirst <> nil
end;

procedure T0Object.FreeObjects;
begin
	while LastObject <> nil do
		LastObject.Free
end;

function T0Object.EnumerateObjects(Proc: CEachObject; Context: Pointer): T0Object;
asm
	push ebp
	push ebx
	push esi
	push edi
	mov ebx, T0Object[eax].FFirst {Current}
	mov esi, ecx {Context}
	mov edi, edx {Proc}
@While:
	test ebx, ebx
	jz @Exit
	mov ebp, T0Object[ebx].FSucc {Ɏ𓾂ĂȂFree}
	push esi
	push ebx
	call edi
	add esp, 8
	test al, al
	jnz @Exit
	mov ebx, ebp
	jmp @While
@Exit:
	mov eax, ebx
	pop edi
	pop esi
	pop ebx
	pop ebp
end;

{ T0CollectionItem }

constructor T0CollectionItem.Create(ACollection: T0Collection; ANext: T0CollectionItem);
begin
	inherited Create;

	FOwner := ACollection;
	Inc(ACollection.FCount);

	FSucc := ANext;
	if ANext <> nil then FPred := ANext.FPred else FPred := FOwner.FLast;

	if FPred <> nil then FPred.FSucc := Self else FOwner.FFirst := Self;
	if FSucc <> nil then FSucc.FPred := Self else FOwner.FLast := Self;
end;

destructor T0CollectionItem.Destroy;
begin
	Dec(T0Collection(Owner).FCount);
	inherited;
end;

{ T0Collection }

procedure T0Collection.Clear;
var
	R, Next: T0Object;
begin
	R := FFirst;
	while R <> nil do
	begin
		Next := R.FSucc;
		if R is T0CollectionItem then R.Free;
		R := Next;
	end;
end;

function T0Collection.Enumerate(ItemClass: TClass; Proc: CEachObject; Context: Pointer): T0Object;
var
	Next: T0Object;
begin
	Result := FFirst;
	while Result <> nil do
	begin
		Next := Result.FSucc;
		if (Result is ItemClass) and Proc(Result, Context) then Exit;
		Result := Next;
	end;
end;

{ T0AggregatedObject }

function T0AggregatedObject._AddRef: Integer;
begin
	if FController <> nil then
		Result := IInterface(FController)._AddRef
	else
		Result := inherited _AddRef
end;

function T0AggregatedObject._Release: Integer;
begin
	if FController <> nil then
		Result := IInterface(FController)._Release
	else
		Result := inherited _Release
end;

function T0AggregatedObject.QueryInterface(const IID: TGUID; out Obj): HResult;
begin
	if FController <> nil then
		Result := IInterface(FController).QueryInterface(IID, Obj)
	else
		Result := inherited QueryInterface(IID, Obj)
end;

procedure T0AggregatedObject.SetController(const Value: IInterface);
begin
	FController := Pointer(Value)
end;

{ TObservable }

procedure TObservable.Attach(const AView: IObserver);
begin
	SetLength(FViews, Length(FViews) + 1);
	FViews[High(FViews)] := AView;

	ViewIncreased;
end;

procedure TObservable.Detach(const AView: IObserver);
var
	I, J: Integer;
begin
	for I := High(FViews) downto Low(FViews) do
		if FViews[I] = AView then
		begin
			for J := I + 1 to High(FViews) do
				FViews[J - 1] := FViews[J];

			SetLength(FViews, Length(FViews) - 1);
			ViewDecreased;
			Break;
		end;
end;

procedure TObservable.Notify;
var
	I: Integer;
	C: Pointer;
begin
	{EĂvȂ悤Ɂc}
	I := Low(FViews);
	while I <= High(FViews) do
	begin
		C := Pointer(FViews[I]);
		FViews[I].Observe;
		if (I > High(FViews)) or (C = Pointer(FViews[I])) then Inc(I);
	end;
end;

procedure TObservable.ViewDecreased;
begin
end;

procedure TObservable.ViewIncreased;
begin
end;

function TObservable.HasClients: Boolean;
begin
	Result := FViews <> nil
end;

end.

