unit LOpenGL;

{*******************************************************}
{                                                       }
{       Light Windows Core                              }
{       OpenGL                                          }
{                                                       }
{       2004 YT                                         }
{                                                       }
{*******************************************************}

interface

uses
	Windows, Messages, SysUtils, Types, dglOpenGL,
	LObjects, LWindows;

type
	T3DScreen = class;

	IScreenCreator = interface(IInterface)
		procedure InitializeScreen;
		procedure FinalizeScreen;
	end;

	T3DScreen = class(T0Window)
	private
		FFullScreenSize: TSize;
		FCreator: IScreenCreator;
		FScreenWindow: HWND;
		FDC: HDC;
		FRC: HGLRC;
		FChanging: Integer;
		FFullScreen: Boolean;
		FStandby: Boolean;
		FCursorVisible: Boolean;
		procedure WMSetFocus(var Message); message WM_SETFOCUS;
		procedure WMLightWindows(var Message: TMessage); message WM_LIGHTWINDOWS;
		procedure SetFullScreen(const Value: Boolean);
	protected
		procedure Initialize; override;
		procedure Finalize; override;
		procedure InitializeGL;
		procedure FinalizeGL;
		property ScreenWindow: HWND read FScreenWindow;
	public
		constructor Create(AParent: TAnyWindow;
			const APosition: TPoint; const ASize: TSize;
			const ACreator: IScreenCreator;
			AFullScreen: Boolean; const AFullScreenSize: TSize);
		procedure Realign; override;
		procedure Present;
		function ScreenSize: TSize;
		property CursorVisible: Boolean read FCursorVisible write FCursorVisible;
		property Standby: Boolean read FStandby;
		property FullScreen: Boolean read FFullScreen write SetFullScreen;
		property FullScreenSize: TSize read FFullScreenSize;
	end;

const
	Size640x480: TSize = (cx: 640; cy: 480);
	Size800x600: TSize = (cx: 800; cy: 600);

implementation

const
	ScreenClassName = 'DirectDrawScreen';

function ScreenProc(Wnd: HWND; Msg: UINT; WP: WParam; LP: LParam): LRESULT; stdcall;
var
	Self: T3DScreen;
begin
	Self := T3DScreen(GetWindowLongW(Wnd, GWL_USERDATA));

	case Msg of
	WM_CREATE:
	begin
		SetWindowLongW(Wnd, GWL_USERDATA, LongInt(PCREATESTRUCTW(LP)^.lpCreateParams));
		Result := 0;
	end;
	WM_ACTIVATE:
	begin
		with Self do
			if FullScreen and (FChanging = 0) then
			begin
				Inc(FChanging);
				try
					if WP = WA_ACTIVE then
					begin
						ShowWindow(Wnd, SW_MAXIMIZE);
						WriteLn('ACTIVE');
						InitializeGL;
					end
					else
					begin
						FinalizeGL;
						ShowWindow(Wnd, SW_MINIMIZE);
					end;
				finally
					Dec(FChanging);
				end;
			end;

		Result := DefWindowProcW(Wnd, Msg, WP, LP);
	end;
	WM_SIZE:
	begin
		with Self do
			if Standby and not FullScreen and
				IsWindowVisible(Wnd) and not IsIconic(Wnd) and (FChanging = 0) then
			begin
				FinalizeGL;
				WriteLn('SIZE');
				InitializeGL;
			end;

		Result := DefWindowProcW(Wnd, Msg, WP, LP);
	end;
	WM_SYSCHAR:
	begin
		if WP = 13 then
		begin
			with Self do FullScreen := not FullScreen;
			Result := 0;
		end
		else
		begin
			Result := DefWindowProcW(Wnd, Msg, WP, LP)
		end;
	end;
	WM_SETCURSOR:
	begin
		if not Self.Standby or (GetActiveWindow <> GetForegroundWindow) then
			Result := DefWindowProcW(Wnd, Msg, WP, LP)
		else
		begin
			if Self.CursorVisible then
				SetCursor(LoadCursor(0, IDC_ARROW))
			else
				SetCursor(0);

			Result := 1;
		end;
	end;
	WM_MOUSEFIRST..WM_MOUSELAST:
	begin
		Result := Self.WindowProc(Msg, WP, LP);
	end;
	WM_SYSCOMMAND:
	begin
		if ((WP = SC_KEYMENU) or (WP = SC_MOUSEMENU) or (WP = SC_CLOSE)) and
			Self.Standby and Self.FullScreen
		then
			Result := 0
		else
			Result := DefWindowProcW(Wnd, Msg, WP, LP);
	end;
	WM_GETMINMAXINFO:
	begin
		Result := DefWindowProcW(Wnd, Msg, WP, LP);
		PMinmaxinfo(Lp)^.ptMaxSize.X := Self.FullScreenSize.cx;
		PMinmaxinfo(Lp)^.ptMaxSize.Y := Self.FullScreenSize.cy + 1;
		PMinmaxinfo(Lp)^.ptMaxPosition.Y := -1;
	end;
	else
		Result := DefWindowProcW(Wnd, Msg, WP, LP);
	end;
end;

procedure RegisterScreenWindow;
var
	WndClass: WNDCLASSEXW;
begin
	if not GetClassInfoExW(MainInstance, ScreenClassName, WndClass) then
	begin
		WndClass.cbSize := SizeOf(WndClass);
		WndClass.style := 0; //CS_SAVEBITS;
		WndClass.lpfnWndProc := @ScreenProc;
		WndClass.cbClsExtra := 0;
		WndClass.cbWndExtra := 0;
		WndClass.hInstance := MainInstance;
		WndClass.hIcon := LoadIconW(MainInstance, 'MAINICON');
		WndClass.hCursor := 0;
		WndClass.hbrBackground := GetStockObject(BLACK_BRUSH);
		WndClass.lpszMenuName := nil;
		WndClass.lpszClassName := ScreenClassName;
		WndClass.hIconSm := 0;
		RegisterClassExW(WndClass);
	end
end;

{ T3DScreen }

constructor T3DScreen.Create(AParent: TAnyWindow; const APosition: TPoint;
	const ASize: TSize; const ACreator: IScreenCreator; AFullScreen: Boolean;
	const AFullScreenSize: TSize);
begin
	inherited Create(AParent, '', APosition, ASize);
	FCreator := ACreator;
	FFullScreen := AFullScreen;
	FFullScreenSize := AFullScreenSize;

	if dglOpenGL.LibHandle = 0 then
	begin
		InitOpenGL;
    ReadExtensions;
	end;
end;

procedure T3DScreen.Finalize;
begin
	FinalizeGL;
	DestroyWindow(FScreenWindow);
	inherited;
end;

procedure T3DScreen.FinalizeGL;
begin
	if not Standby then Exit;

	if FCreator <> nil then
		FCreator.FinalizeScreen;

	wglMakeCurrent(FDC, CDS_RESET);
	wglDeleteContext(FRC);
	ReleaseDC(ScreenWindow, FDC);

	if FullScreen then
	begin
		if not CursorVisible then ShowCursor(True);
		ChangeDisplaySettingsW(TDeviceModeW(nil^), 0);
	end;

	FStandby := False;
end;

procedure T3DScreen.Initialize;
var
	S: DWORD;
	P: HWND;
begin
	inherited;

	S := WS_VISIBLE or WS_CHILD;
	P := Handle;
	if FullScreen then
	begin
		S := WS_POPUP or WS_SYSMENU;
		P := 0;
	end;

	RegisterScreenWindow;
	FScreenWindow := CreateWindowExW(WS_EX_TOPMOST or WS_EX_APPWINDOW,
						ScreenClassName, '', S,
						0, 0, 0, 0, P, 0, MainInstance, Self);

	PostMessageW(WM_LIGHTWINDOWS, LW_READY, 0);
end;

procedure T3DScreen.InitializeGL;

	var DisplayBits: Cardinal;

	function SelectDisplayMode: Boolean;
	var
		DC: HDC;
		DeviceMode: TDeviceModeW;
	begin
		DC := GetDC(0);
		DisplayBits := GetDeviceCaps(DC, BITSPIXEL) * GetDeviceCaps(DC, PLANES);
		ReleaseDC(0, DC);

		FillChar(DeviceMode, SizeOf(DeviceMode), 0);
		DeviceMode.dmSize := SizeOf(DeviceMode);
		DeviceMode.dmBitsPerPel := DisplayBits;
		DeviceMode.dmPelsWidth := FullScreenSize.cx;
		DeviceMode.dmPelsHeight := FullScreenSize.cy;
		DeviceMode.dmFields := DM_PELSWIDTH or DM_PELSHEIGHT or DM_BITSPERPEL;
		Result := ChangeDisplaySettingsW(DeviceMode, CDS_FULLSCREEN) = DISP_CHANGE_SUCCESSFUL;
	end;

	function InitOpenGL: Boolean;
	const
		DefaultPixelFormat: Windows.TPixelFormatDescriptor = (
			nSize: Sizeof(Windows.TPixelFormatDescriptor);
			nVersion: 1;
			dwFlags: PFD_DRAW_TO_WINDOW or PFD_SUPPORT_OPENGL or PFD_DOUBLEBUFFER;
			iPixelType: PFD_TYPE_RGBA;
			cColorBits: 24;
			cRedBits: 0;
			cRedShift: 0;
			cGreenBits: 0;
			cGreenShift: 0;
			cBlueBits: 0;
			cBlueShift: 0;
			cAlphaBits: 0;
			cAlphaShift: 0;
			cAccumBits: 0;
			cAccumRedBits: 0;
			cAccumGreenBits: 0;
			cAccumBlueBits: 0;
			cAccumAlphaBits: 0;
			cDepthBits: 32;
			cStencilBits: 0;
			cAuxBuffers: 0;
			iLayerType: PFD_MAIN_PLANE;
			bReserved: 0;
			dwLayerMask: 0;
			dwVisibleMask: 0;
			dwDamageMask: 0;
		);
	var
		PixelFormat: Windows.TPixelFormatDescriptor;
	begin
		PixelFormat := DefaultPixelFormat;
		PixelFormat.cColorBits := DisplayBits;

		FDC := GetDC(ScreenWindow);

		Result := SetPixelFormat(FDC, ChoosePixelFormat(FDC, @PixelFormat), @PixelFormat);
		if not Result then WriteLn('SetPixelFormat');
		if Result then
		begin
			FRC := wglCreateContext(FDC);
			if FRC = 0 then WriteLn('wglCreateContext');
			Result := FRC <> 0;
			if Result then
			begin
				Result := wglMakeCurrent(FDC, FRC);
				if not Result then WriteLn('wglMakeCurrent');
			end
		end;
	end;

var
	R: TRect;
begin
	if Standby then Exit;

	if not FullScreen or SelectDisplayMode then
	begin
		if InitOpenGL then
		begin
			if FullScreen then
			begin
				GetWindowRect(ScreenWindow, R);
				//glViewport(0, R.Bottom - FFullScreenSize.cy, FFullScreenSize.cx, FFullScreenSize.cy);
				glViewport(0, 0, FFullScreenSize.cx, FFullScreenSize.cy);
				if not CursorVisible then ShowCursor(False);
			end;

			FStandby := True;
			if Standby and (FCreator <> nil) then
				FCreator.InitializeScreen;
		end;
	end;
end;

procedure T3DScreen.Present;
begin
	SwapBuffers(FDC);
end;

procedure T3DScreen.Realign;
var
	Size: TSize;
begin
	Size := ClientSize;
	MoveWindow(FScreenWindow, 0, 0, Size.cx, Size.cy, False);
end;

function T3DScreen.ScreenSize: TSize;
var
	R: TRect;
begin
	if FullScreen then
		Result := FFullScreenSize
	else
	begin
		GetWindowRect(FScreenWindow, R);
		Result.cx := R.Right - R.Left;
		Result.cy := R.Bottom - R.Top
	end
end;

procedure T3DScreen.SetFullScreen(const Value: Boolean);

	procedure ReadyFullScreen;
	var
		Info: MENUITEMINFOW;
	begin
		ShowWindow(FScreenWindow, SW_HIDE);
		Windows.SetParent(FScreenWindow, 0);
		SetWindowLongW(FScreenWindow, GWL_STYLE, Integer(WS_POPUP or WS_SYSMENU));

		Info.cbSize := SizeOf(Info);
		Info.fMask := MIIM_STATE;
		Info.fState := MFS_DISABLED;
		SetMenuItemInfoW(GetSystemMenu(FScreenWindow, False), SC_CLOSE, False, Info);

		ShowWindow(FScreenWindow, SW_MAXIMIZE);
		ShowWindow(TopParent.Handle, SW_HIDE);
	end;

	procedure RestoreFullScreen;
	begin
		ShowWindow(FScreenWindow, SW_HIDE);
		SetWindowLongW(FScreenWindow, GWL_STYLE, WS_CHILD);
		Windows.SetParent(FScreenWindow, Handle);
		Realign;
		ShowWindow(FScreenWindow, SW_SHOW);
		ShowWindow(TopParent.Handle, SW_SHOW);
	end;

var
	OldStandby: Boolean;
begin
	if Value = FFullScreen then Exit;
	OldStandby := Standby;

	Inc(FChanging);
	try
		FinalizeGL;

		FFullScreen := Value;
		if Value then
			ReadyFullScreen
		else
			RestoreFullScreen;

		WriteLn('FULL');
		InitializeGL;
	finally
		Dec(FChanging);
	end;

	if not Standby and OldStandby then
		SetFullScreen(not Value);
end;

procedure T3DScreen.WMLightWindows(var Message: TMessage);
begin
	if Message.WParam = LW_READY then
	begin
		Inc(FChanging);
		try
			if FullScreen then
			begin
				TopParent.Hide;
				ShowWindow(ScreenWindow, SW_MAXIMIZE);
			end;
			InitializeGL;
		finally
			Dec(FChanging)
		end;
	end
	else
		inherited
end;

procedure T3DScreen.WMSetFocus(var Message);
begin
	Windows.SetFocus(FScreenWindow);
end;

end.

