unit LRegistry;

{*******************************************************}
{                                                       }
{       Light Windows                                   }
{       Registry Access                                 }
{                                                       }
{       2000..2004, 2008 YT                             }
{                                                       }
{*******************************************************}

interface

uses
	Windows, SysUtils, SysConst,
	SysUtilsW, SysHacks, AlgUtils, WinTips,
	LObjects, LConsts, LStreams;

const
	RegKeyAll  = KEY_ALL_ACCESS;
	RegKeyRead = KEY_READ;
	RegKeyPath = KEY_READ or KEY_CREATE_SUB_KEY;

type
	TRegKeyInfo = record
		NumSubKeys:   DWORD;
		MaxSubKeyLen: DWORD;
		NumValues:    DWORD;
		MaxValueLen:  DWORD;
		MaxDataLen:   DWORD;
		FileTime:     Windows.FILETIME;
	end;

	TRegValueInfo = record
		DataType: Cardinal;
		DataSize: DWORD;
	end;

	CEachName = function(const Name: WideString; Context: Pointer): Boolean cdecl;

	T0Registry = class(T0Object)
	private
		FHandle: HKEY;
	public
		constructor Create(AHandle: HKEY);
		function GetKeyInfo(var Info: TRegKeyInfo): Boolean;
		function HasSubKeys: Boolean;
		function KeyExists(const Name: WideString): Boolean;
		procedure EnumerateKeys(Proc: CEachName; Context: Pointer);
		property Handle: HKEY read FHandle;
	end;

	TRegistryHive = class(T0Registry)
	public
		constructor Create(AHandle: HKEY);
	end;

	TRegistryKey = class(T0Registry)
	private
		FAccess: LongWord;
		FLazyWrite: Boolean;
	protected
		function GetData(const Name: WideString; Buffer: Pointer; BufSize: Cardinal; var DataType: Cardinal): Integer;
		procedure PutData(const Name: WideString; Buffer: Pointer; BufSize: Cardinal; DataType: Cardinal);
	public
		constructor Create(APath: T0Registry; const AName: WideString;
						AAccess: LongWord = KEY_READ; CanCreate: Boolean = False); overload;
		constructor Create(AParent: HKEY; const AName: WideString;
						AAccess: LongWord = KEY_READ; CanCreate: Boolean = False); overload;
		procedure BeforeDestruction; override;
		function GetValueInfo(const Name: WideString; var Info: TRegValueInfo): Boolean;
		function ValueExists(const Name: WideString): Boolean;
		function HasValues: Boolean;
		function Read: WideString;
		function ReadString(const Name: WideString): WideString;
		{ReadStringExpandStringǂނƊϐ̓WJs}
		function ReadExpandString(const Name: WideString): WideString;
		procedure WriteString(const Name, Value: WideString);
		procedure WriteExpandString(const Name, Value: WideString);
		function ReadInteger(const Name: WideString): Integer;
		procedure WriteInteger(const Name: WideString; Value: Integer);
		function ReadBool(const Name: WideString): Boolean;
		procedure WriteBool(const Name: WideString; Value: Boolean);
		function ReadBinaryData(const Name: WideString;
						var Buffer;	BufSize: Cardinal): Integer;
		procedure WriteBinaryData(const Name: WideString; var Buffer; BufSize: Cardinal);
		procedure EnumerateValues(Proc: CEachName; Context: Pointer);
		property Access: LongWord read FAccess;
		property LazyWrite: Boolean read FLazyWrite write FLazyWrite;
	end;

type
	TIniFileValueRec = record
		NameHash: THashCode;
		Name: WideString;
		Value: WideString;
	end;

	PIniFileSectionRec = ^TIniFileSectionRec;
	TIniFileSectionRec = record
		Pred: PIniFileSectionRec;
		Succ: PIniFileSectionRec;
		Name: WideString;
		Values: array of TIniFileValueRec;
	end;

	TIniFileSection = class;

	TIniFile = class(T0Object)
	private
		FFileName: TWideFileName;
		FFirst: PIniFileSectionRec;
		FLast: PIniFileSectionRec;
		FModified: Boolean;
		function GetSections(const AName: WideString): TIniFileSection;
	protected
		function NewSection(const AName: WideString): PIniFileSectionRec;
		procedure DisposeSection(ASection: PIniFileSectionRec);
		function GetSection(const AName: WideString): PIniFileSectionRec;
		procedure Load;
		procedure Save;
	public
		constructor Create(const AFileName: TWideFileName);
		procedure BeforeDestruction; override;
		property FileName: TWideFileName read FFileName;
		property Sections[const AName: WideString]: TIniFileSection read GetSections;
	end;

	TIniFileSection = class(T0Object)
	private
		FIniFile: TIniFile;
		FSection: PIniFileSectionRec;
		function GetName: WideString;
	public
		constructor Create(AIniFile: TIniFile; const AName: WideString);
		function ReadString(const AName: WideString): WideString;
		procedure WriteString(const AName, AValue: WideString);
		property Name: WideString read GetName;
	end;

type
	EStorage = class(Exception);
	ERegistryError = class(EStorage);
	EIniFileError = class(EStorage);

const
	HiveTexts: array[HKEY_CLASSES_ROOT..HKEY_DYN_DATA] of WideString = (
		'HKEY_CLASSES_ROOT',
		'HKEY_CURRENT_USER',
		'HKEY_LOCAL_MACHINE',
		'HKEY_USERS',
		'HKEY_PERFORMANCE_DATA',
		'HKEY_CURRENT_CONFIG',
		'HKEY_DYN_DATA');

implementation

{ T0Registry }

constructor T0Registry.Create(AHandle: HKEY);
begin
	inherited Create;
	FHandle := AHandle;
end;

procedure T0Registry.EnumerateKeys(Proc: CEachName; Context: Pointer);
var
	Info: TRegKeyInfo;
	Len: DWORD;
	Temp: WideString;
	Index: Integer;
begin
	if not GetKeyInfo(Info) then
		raise ERegistryError.Create(SNoAllowed);

	for Index := 0 to Info.NumSubKeys - 1 do
	begin
		Len := Info.MaxSubKeyLen + 1;
		SetString(Temp, nil, Len);
		RegEnumKeyExW(Handle, Index, PWideChar(Temp), Len, nil, nil, nil, nil);
		if Proc(PWideChar(Temp), Context) then Break;
	end;
end;

function T0Registry.GetKeyInfo(var Info: TRegKeyInfo): Boolean;
begin
	FillChar(Info, SizeOf(Info), 0);
	Result := RegQueryInfoKeyW(Handle, nil, nil, nil, @Info.NumSubKeys,
					@Info.MaxSubKeyLen, nil, @Info.NumValues, @Info.MaxValueLen,
					@Info.MaxDataLen, nil, @Info.FileTime) = ERROR_SUCCESS;
{
	if SysLocale.FarEast and (Win32Platform = VER_PLATFORM_WIN32_NT) then
	begin
		Inc(Info.MaxSubKeyLen, Info.MaxSubKeyLen); //Unicode 
		Inc(Info.MaxValueLen, Info.MaxValueLen);
	end;
}
end;

function T0Registry.HasSubKeys: Boolean;
var
	Info: TRegKeyInfo;
begin
	Result := GetKeyInfo(Info) and (Info.NumSubKeys > 0);
end;

function T0Registry.KeyExists(const Name: WideString): Boolean;
var
	TempKey: HKEY;
begin
	Result := RegOpenKeyExW(Handle, PWideChar(Name), 0, KEY_READ, TempKey) = ERROR_SUCCESS;
	if Result then
		RegCloseKey(TempKey);
end;

{ TRegistryHive }

constructor TRegistryHive.Create(AHandle: HKEY);
begin
	Assert((AHandle >= Low(HiveTexts)) and (AHandle <= High(HiveTexts)));
	inherited Create(AHandle);
end;

{ TRegistryKey }

constructor TRegistryKey.Create(AParent: HKEY; const AName: WideString;
	AAccess: LongWord; CanCreate: Boolean);
var
	AHandle: HKEY;
	Disposition: DWORD;
	R: DWORD;
begin
	FAccess := AAccess;
	FLazyWrite := True;

	if not CanCreate then
		R := RegOpenKeyExW(AParent, PWideChar(AName), 0, AAccess, AHandle)
	else
		R := RegCreateKeyExW(AParent, PWideChar(AName), 0, nil,
						REG_OPTION_NON_VOLATILE, FAccess, nil, AHandle, @Disposition);

	if R <> ERROR_SUCCESS then
		raise ERegistryError.CreateFmt(SFailedOpen, [AName]);

	inherited Create(AHandle);
end;

constructor TRegistryKey.Create(APath: T0Registry; const AName: WideString;
				AAccess: LongWord; CanCreate: Boolean);
begin
	Create(APath.Handle, AName, AAccess, CanCreate);
end;

procedure TRegistryKey.BeforeDestruction;
begin
	{̃`FbN̓RXgN^ŗÔ}
	if Handle <> 0 then
	begin
		{tbV}
		if not LazyWrite then
			RegFlushKey(Handle);

		{}
		RegCloseKey(Handle);
	end;
	inherited;
end;

function TRegistryKey.
				GetValueInfo(const Name: WideString; var Info: TRegValueInfo): Boolean;
begin
	FillChar(Info, SizeOf(Info), 0);
	{NamenilłPointerLXg}
	Result := RegQueryValueExW(Handle, Pointer(Name), nil,
					@Info.DataType, nil, @Info.DataSize) = ERROR_SUCCESS;
end;

function TRegistryKey.GetData(const Name: WideString;
				Buffer: Pointer; BufSize: Cardinal; var DataType: Cardinal): Integer;
begin
	if RegQueryValueExW(Handle, PWideChar(Name), nil,
					@DataType, Buffer, @BufSize) <> ERROR_SUCCESS then
		raise ERegistryError.CreateFmt(SFailedGet, [Name]);

	Result := BufSize;
end;

function TRegistryKey.ValueExists(const Name: WideString): Boolean;
var
	Info: TRegValueInfo;
begin
	Result := GetValueInfo(Name, Info);
end;

procedure TRegistryKey.PutData(const Name: WideString;
				Buffer: Pointer; BufSize: Cardinal; DataType: Cardinal);
begin
	if RegSetValueExW(Handle, PWideChar(Name), 0,
					DataType, Buffer, BufSize) <> ERROR_SUCCESS then
		raise ERegistryError.CreateFmt(SFailedSet, [Name]);
end;

function TRegistryKey.ReadString(const Name: WideString): WideString;
var
	D: TRegValueInfo;
	R: Boolean;
begin
	R := GetValueInfo(Name, D);
	if not R or not (D.DataType in [REG_SZ, REG_EXPAND_SZ]) then
	begin
		if Name = '' then
			Result := ''
		else
			raise ERegistryError.CreateFmt(SInvalidDataType, [Name]);
	end
	else if D.DataSize <= 1 then
		Result := ''
	else
	begin
		SetLength(Result, (D.DataSize div SizeOf(WideChar)) - 1);
		GetData(Name, Pointer(Result), D.DataSize, D.DataType);
		if D.DataType = REG_EXPAND_SZ then
			Result := ExpandEnvironmentStrings(Result);
	end;
end;

procedure TRegistryKey.WriteString(const Name, Value: WideString);
begin
	PutData(Name, PWideChar(Value), (Length(Value) + 1) * SizeOf(WideChar), REG_SZ);
end;

function TRegistryKey.ReadExpandString(const Name: WideString): WideString;
var
	D: TRegValueInfo;
	R: Boolean;
begin
	R := GetValueInfo(Name, D);
	if not R or (D.DataType <> REG_EXPAND_SZ) then
		raise ERegistryError.CreateFmt(SInvalidDataType, [Name])
	else if D.DataSize <= 1 then
		Result := ''
	else
	begin
		SetLength(Result, (D.DataSize div SizeOf(WideChar)) - 1);
		GetData(Name, Pointer(Result), D.DataSize, D.DataType);
	end;
end;

procedure TRegistryKey.WriteExpandString(const Name, Value: WideString);
begin
	PutData(Name, PWideChar(Value), (Length(Value) + 1) * SizeOf(WideChar), REG_EXPAND_SZ);
end;

function TRegistryKey.ReadInteger(const Name: WideString): Integer;
var
	RegData: Cardinal;
begin
	GetData(Name, @Result, SizeOf(Result), RegData);
	if RegData <> REG_DWORD then
		raise ERegistryError.CreateFmt(SInvalidDataType, [Name])
end;

procedure TRegistryKey.WriteInteger(const Name: WideString; Value: Integer);
begin
	PutData(Name, @Value, SizeOf(Value), REG_DWORD);
end;

function TRegistryKey.ReadBool(const Name: WideString): Boolean;
var
	Buffer: array[0..4] of WideChar;
	RegData: Cardinal;
begin
	GetData(Name, @Buffer, SizeOf(Buffer), RegData);
	case RegData of
	REG_DWORD:
		Result := PInteger(@Buffer)^ <> 0; {̂?}
	REG_SZ:
		Result := (lstrcmpiW(Buffer, '1') = 0) or
						(lstrcmpiW(Buffer, 'True') = 0) or
						(lstrcmpiW(Buffer, 'Yes') = 0) or
						(lstrcmpiW(Buffer, 'On') = 0);
	else
		raise ERegistryError.CreateFmt(SInvalidDataType, [Name])
	end;
end;

procedure TRegistryKey.WriteBool(const Name: WideString; Value: Boolean);
begin
	WriteInteger(Name, Ord(Value));
end;

function TRegistryKey.ReadBinaryData(const Name: WideString;
				var Buffer; BufSize: Cardinal): Integer;
var
	Info: TRegValueInfo;
	R: Boolean;
begin
	R:= GetValueInfo(Name, Info);
	if not R or (Info.DataType <> REG_BINARY) or (Info.DataSize > BufSize) then
		raise ERegistryError.CreateFmt(SInvalidDataType, [Name]);

	Result := Info.DataSize;
	GetData(Name, @Buffer, Result, Info.DataType)
end;

procedure TRegistryKey.WriteBinaryData(const Name: WideString; var Buffer; BufSize: Cardinal);
begin
	PutData(Name, @Buffer, BufSize, REG_BINARY);
end;

procedure TRegistryKey.EnumerateValues(Proc: CEachName; Context: Pointer);
var
	Info: TRegKeyInfo;
	Len: DWORD;
	Temp: WideString;
	Index: Integer;
begin
	if not GetKeyInfo(Info) then
		raise ERegistryError.Create(SNoAllowed);

	for Index := 0 to Info.NumValues - 1 do
	begin
		Len := Info.MaxValueLen + 1;
		SetString(Temp, nil, Len);
		RegEnumValueW(Handle, Index, Pointer(PWideChar(Temp)), Len, nil, nil, nil, nil);
		if Proc(PWideChar(Temp), Context) then Break;
	end;
end;

function TRegistryKey.HasValues: Boolean;
var
	Info: TRegKeyInfo;
begin
	Result := GetKeyInfo(Info) and (Info.NumValues > 0)
end;

function TRegistryKey.Read: WideString;
begin
	Result := ReadString('')
end;

{ TIniFile }

constructor TIniFile.Create(const AFileName: TWideFileName);
begin
	inherited Create;
	FFileName := AFileName;
	if FileExists(AFileName) then Load;
end;

procedure TIniFile.BeforeDestruction;
begin
	if FModified then Save;
	while FFirst <> nil do DisposeSection(FFirst);
	inherited;
end;

procedure TIniFile.DisposeSection(ASection: PIniFileSectionRec);
begin
	RemoveFromLinkList(FFirst, FLast, ASection, ASection^.Pred, ASection^.Succ);
	Dispose(ASection)
end;

function TIniFile.GetSections(const AName: WideString): TIniFileSection;

	function FindProc(Obj: T0Object): Boolean cdecl;
	begin
		Result := (Obj is TIniFileSection) and
						WideSameText(TIniFileSection(Obj).Name, AName)
	end;

begin
	Result := TIniFileSection(EnumerateObjects(CEachObject(@FindProc), GetContext));
	if Result <> nil then Exit;

	Result := TIniFileSection.Create(Self, AName);
	Result.Owner := Self
end;

function TIniFile.GetSection(const AName: WideString): PIniFileSectionRec;
var
	I: PIniFileSectionRec;
begin
	I := FFirst;
	while I <> nil do
	begin
		if WideSameText(I^.Name, AName) then
		begin
			Result := I;
			Exit;
		end;
		I := I.Succ;
	end;
	Result := NewSection(AName);
end;

function TIniFile.NewSection(const AName: WideString): PIniFileSectionRec;
begin
	New(Result);
	Result^.Name := AName;
	AddToLinkList(FFirst, FLast, Result, Result^.Pred, Result^.Succ);
end;

procedure TIniFile.Load;
var
	I: PIniFileSectionRec;
	S, N: WideString;
	H: TFileStream;
	P: Integer;
begin
	H := TFileStream.Create(FileName, fmOpenRead);
	try
		I := nil;
		while H.Rest > 0 do
		begin
			S := Trim(UTF8Decode(ReadLineFromStream(H)));
			if (S = '') or (S[1] = ';') then
				begin	end
			else if (S[1] = '[') and (S[Length(S)] = ']') then
			begin
				I := NewSection(Trim(Copy(S, 2, Length(S) - 2)))
			end
			else
			begin
				if I = nil then
					raise EIniFileError.CreateFmt(SInvalidFormat, [FileName]);

				P := Pos('=', S);
				if P = 0 then P := Length(S);

        N := Trim(Copy(S, 1, P - 1));

				SetLength(I^.Values, Length(I^.Values) + 1);
				with I^.Values[High(I^.Values)] do
				begin
					NameHash := GetHashCode(N);
					Name := N;
					Value := Trim(Copy(S, P + 1, MaxInt))
				end	
			end
		end;
	finally
		H.Free
	end
end;

procedure TIniFile.Save;
var
	I: PIniFileSectionRec;
	J: Integer;
	H: TFileStream;
begin
	H := TFileStream.Create(FileName, fmCreate);
	try
		I := FFirst;
		while I <> nil do
		begin
			WriteLineToStream(H, UTF8Encode('[' + I^.Name + ']'));
			for J := Low(I^.Values) to High(I^.Values) do
				WriteLineToStream(H, UTF8Encode(I^.Values[J].Name + '=' + I^.Values[J].Value));
			WriteLineToStream(H);
			I := I.Succ;
		end;
	finally
		H.Free
	end
end;

{ TIniFileSection }

constructor TIniFileSection.Create(AIniFile: TIniFile; const AName: WideString);
begin
	inherited Create;
	FIniFile := AIniFile;
	FSection := AIniFile.GetSection(AName);
end;

function TIniFileSection.GetName: WideString;
begin
	Result := FSection^.Name
end;

function TIniFileSection.ReadString(const AName: WideString): WideString;
var
	Hash: THashCode;
	I: Integer;
begin
	Hash := GetHashCode(AName);
	for I := Low(FSection.Values) to High(FSection.Values) do
		if FSection.Values[I].NameHash = Hash then
			if FSection.Values[I].Name = AName then
			begin
				Result := FSection.Values[I].Value;
				Exit;
			end;

	Result := ''
end;

procedure TIniFileSection.WriteString(const AName, AValue: WideString);
var
	Hash: THashCode;
	I: Integer;
begin
	Hash := GetHashCode(AName);
	for I := Low(FSection.Values) to High(FSection.Values) do
		if FSection.Values[I].NameHash = Hash then
			if FSection.Values[I].Name = AName then
			begin
				if FSection.Values[I].Value <> AValue then
				begin
					FSection.Values[I].Value := AValue;
					FIniFile.FModified := True;
				end;
				Exit;
			end;

	SetLength(FSection.Values, Length(FSection.Values) + 1);

	with FSection.Values[High(FSection.Values)] do
	begin
		NameHash := Hash;
		Name := AName;
		Value := AValue;
	end;
	FIniFile.FModified := True;
end;

end.

